# Copyright 2004-2009 Joe Wreschnig, Michael Urman, Iñigo Serna,
#                     Steven Robertson
#           2011-2023 Nick Boultbee
#           2013      Christoph Reiter
#           2014      Jan Path
#           2023      Jej@github
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

from gi.repository import Gtk, Gio

from quodlibet import config
from quodlibet import qltk
from quodlibet import util
from quodlibet import app
from quodlibet import C_, _
from quodlibet.config import RATINGS, DurationFormat, DURATION

from quodlibet.qltk.ccb import ConfigSwitch as CS
from quodlibet.qltk.data_editors import TagListEditor
from quodlibet.qltk.entry import ValidatingEntry, UndoEntry, ClearEntry
from quodlibet.query._query import Query
from quodlibet.qltk.scanbox import ScanBox
from quodlibet.qltk.maskedbox import MaskedBox
from quodlibet.qltk.songlist import SongList, get_columns
from quodlibet.qltk.window import UniqueWindow
from quodlibet.qltk.x import Button, Align
from quodlibet.qltk.advanced_prefs import AdvancedPreferencesPane
from quodlibet.qltk import Icons, add_css
from quodlibet.util import copool, format_time_preferred
from quodlibet.util.dprint import print_d
from quodlibet.util.library import emit_signal, get_scan_dirs, scan_library
from quodlibet.util import connect_obj

MARGIN = 12

TOP_MARGIN = 3


class PreferencesWindow(UniqueWindow):
    """The tabbed container window for the main preferences GUI.
    Individual tabs are encapsulated as inner classes inheriting from `VBox`"""

    class SongList(Gtk.VBox):
        name = "songlist"

        PREDEFINED_TAGS = [
            ("~#disc", _("_Disc")),
            ("~#track", _("_Track")),
            ("grouping", _("Grou_ping")),
            ("artist", _("_Artist")),
            ("album", _("Al_bum")),
            ("title", util.tag("title")),
            ("genre", _("_Genre")),
            ("date", _("_Date")),
            ("~basename", _("_Filename")),
            ("~#length", _("_Length")),
            ("~rating", _("_Rating")),
            ("~#filesize", util.tag("~#filesize")),
        ]

        def __init__(self):
            def create_behavior_frame():
                vbox = Gtk.VBox(spacing=12)
                jump_button = CS(
                    _("_Jump to playing song automatically"),
                    "settings",
                    "jump",
                    populate=True,
                    tooltip=_(
                        "When the playing song changes, "
                        "scroll to it in the song list"
                    ),
                )
                autosort_button = CS(
                    _("_Sort songs when tags are modified"),
                    "song_list",
                    "auto_sort",
                    populate=True,
                    tooltip=_(
                        "Automatically re-sort songs in "
                        "the song list when tags are modified"
                    ),
                )
                always_sortable = CS(
                    _("Always allow sorting"),
                    "song_list",
                    "always_allow_sorting",
                    populate=True,
                    tooltip=_(
                        "Allow sorting by column headers, even for playlists etc"
                    ),
                )

                def refresh_browser(*args):
                    app.window.set_sortability()

                always_sortable.connect("notify::active", refresh_browser)
                vbox.pack_start(jump_button, False, True, 0)
                vbox.pack_start(always_sortable, False, True, 0)
                vbox.pack_start(autosort_button, False, True, 0)
                return qltk.Frame(_("Behavior"), child=vbox)

            def create_visible_columns_widgets():
                buttons = {}
                vbox = Gtk.VBox(spacing=12)
                grid = Gtk.FlowBox(column_spacing=24)
                for _i, (k, t) in enumerate(self.PREDEFINED_TAGS):
                    buttons[k] = Gtk.CheckButton(label=t, use_underline=True)
                    grid.add(buttons[k])
                vbox.pack_start(grid, False, True, 0)
                # Other columns
                hbox = Gtk.HBox(spacing=12)
                l = Gtk.Label(label=_("_Others:"), use_underline=True)
                hbox.pack_start(l, False, True, 0)
                self.others = others = UndoEntry()
                others.set_sensitive(False)
                # Stock edit doesn't have ellipsis chars.
                edit_button = Button(_("_Edit…"), Icons.EDIT)
                edit_button.connect("clicked", self.__config_cols, buttons)
                edit_button.set_tooltip_text(
                    _("Add or remove additional column headers")
                )
                l.set_mnemonic_widget(edit_button)
                l.set_use_underline(True)
                hbox.pack_start(others, True, True, 0)
                vbox.pack_start(hbox, False, True, 0)
                b = Gtk.HButtonBox()
                b.set_layout(Gtk.ButtonBoxStyle.END)
                b.pack_start(edit_button, True, True, 0)
                vbox.pack_start(b, True, True, 0)
                return qltk.Frame(_("Visible Columns"), child=vbox), buttons

            def create_columns_prefs_frame():
                tiv = Gtk.Switch()
                aio = Gtk.Switch()
                aip = Gtk.Switch()
                fip = Gtk.Switch()
                self._toggle_data = [
                    (tiv, "title", "~title~version"),
                    (aip, "album", "~album~discsubtitle"),
                    (fip, "~basename", "~filename"),
                    (aio, "artist", "~people"),
                ]

                def pack_with_label(vb: Gtk.Box, widget: Gtk.Widget, text: str):
                    hb = Gtk.Box(spacing=12)
                    label = Gtk.Label(label=text, use_underline=True)
                    hb.pack_start(label, False, False, 0)
                    hb.pack_end(widget, False, False, 0)
                    vb.pack_start(hb, False, False, 0)

                vb = Gtk.VBox(spacing=12)
                pack_with_label(vb, tiv, _("Title includes _version"))
                pack_with_label(vb, aip, _("Album includes _disc subtitle"))
                pack_with_label(vb, aio, _("Artist includes all _people"))
                pack_with_label(vb, fip, _("Filename includes _folder"))
                return qltk.Frame(_("Column Preferences"), child=vb)

            def create_update_columns_button():
                apply = Button(_("_Update Columns"), Icons.VIEW_REFRESH)
                apply.set_tooltip_text(
                    _(
                        "Apply current configuration to song list, "
                        "adding new columns to the end"
                    )
                )
                apply.connect("clicked", self.__apply, buttons)
                # Apply on destroy, else config gets mangled
                self.connect("destroy", self.__apply, buttons)
                b = Gtk.HButtonBox()
                b.set_layout(Gtk.ButtonBoxStyle.END)
                b.pack_start(apply, True, True, 0)
                return b

            super().__init__(spacing=12)
            # Store ordered columns
            self._columns = []
            self.set_border_width(12)
            self.title = _("Song List")
            visible_columns_frame, buttons = create_visible_columns_widgets()
            self.pack_start(create_behavior_frame(), False, True, TOP_MARGIN)
            self.pack_start(visible_columns_frame, False, True, MARGIN)
            self.pack_start(create_columns_prefs_frame(), False, True, MARGIN)
            self.pack_start(create_update_columns_button(), False, False, 0)

            # Run it now
            self.__update(buttons, self._toggle_data, get_columns())

            for child in self.get_children():
                child.show_all()

        def __update(self, buttons, toggle_data, columns):
            """Updates all widgets based on the passed column list"""
            self._columns = columns
            columns = list(columns)

            for key, widget in buttons.items():
                widget.set_active(key in columns)
                if key in columns:
                    columns.remove(key)

            for check, off, on in toggle_data:
                if on in columns:
                    buttons[off].set_active(True)
                    check.set_active(True)
                    columns.remove(on)

            self.others.set_text(", ".join(columns))
            self.other_cols = columns

        def __get_current_columns(self, buttons):
            """Given the current column list and the widgets states compute
            a new column list.
            """
            new_headers = set()
            # Get the checked headers
            for key, _name in self.PREDEFINED_TAGS:
                if buttons[key].get_active():
                    new_headers.add(key)
                # And the customs
            new_headers.update(set(self.other_cols))

            on_to_off = {on: off for (w, off, on) in self._toggle_data}
            result = []
            cur_cols = get_columns()
            for h in cur_cols:
                if h in new_headers:
                    result.append(h)
                else:
                    try:
                        alternative = on_to_off[h]
                        if alternative in new_headers:
                            result.append(alternative)
                    except KeyError:
                        pass

            # Add new ones, trying to preserve order
            for new in new_headers - set(result):
                try:
                    idx = self._columns.index(new)
                except ValueError:
                    idx = len(self._columns)
                result.insert(idx, new)

            # After this, do the substitutions
            for check, off, on in self._toggle_data:
                if check.get_active():
                    try:
                        result[result.index(off)] = on
                    except ValueError:
                        pass

            return result

        def __apply(self, button, buttons):
            result = self.__get_current_columns(buttons)
            SongList.set_all_column_headers(result)

        def __config_cols(self, button, buttons):
            def __closed(widget):
                self.__update(buttons, self._toggle_data, widget.tags)

            columns = self.__get_current_columns(buttons)
            m = TagListEditor(_("Edit Columns"), columns)
            m.set_transient_for(qltk.get_top_parent(self))
            m.connect("destroy", __closed)
            m.show()

    class Browsers(Gtk.VBox):
        name = "browser"

        def __init__(self):
            def create_display_frame():
                vbox = Gtk.VBox(spacing=MARGIN)
                model = Gtk.ListStore(str, str)

                def on_changed(combo):
                    it = combo.get_active_iter()
                    if it is None:
                        return
                    DURATION.format = model[it][0]
                    app.window.songlist.info.refresh()
                    app.window.qexpander.refresh()
                    # TODO: refresh info windows ideally too (but see #2019)

                def draw_duration(column, cell, model, it, data):
                    df, example = model[it]
                    cell.set_property("text", example)

                for df in sorted(DurationFormat.values):
                    # 4954s == longest ever CD, FWIW
                    model.append([df, format_time_preferred(4954, df)])
                duration = Gtk.ComboBox(model=model)
                cell = Gtk.CellRendererText()
                duration.pack_start(cell, True)
                duration.set_cell_data_func(cell, draw_duration, None)
                index = sorted(DurationFormat.values).index(DURATION.format)
                duration.set_active(index)
                duration.connect("changed", on_changed)
                hbox = Gtk.HBox(spacing=MARGIN)
                label = Gtk.Label(label=_("Duration totals") + ":", use_underline=True)
                label.set_mnemonic_widget(duration)
                hbox.pack_start(label, False, True, 0)
                hbox.pack_start(duration, False, True, 0)

                vbox.pack_start(hbox, False, True, 0)
                return qltk.Frame(_("Display"), child=vbox)

            def create_search_frame():
                vb = Gtk.VBox(spacing=MARGIN)
                l = Gtk.Label(label=_("_Global filter:"))
                l.set_use_underline(True)
                e = ValidatingEntry(Query.validator)
                e.set_text(config.get("browsers", "background"))
                e.connect("changed", self._entry, "background", "browsers")
                e.set_tooltip_text(_("Apply this query in addition to all others"))
                l.set_mnemonic_widget(e)
                vb.pack_start(hbox_for(l, e), False, True, 0)
                # Translators: The heading of the preference group, no action
                return qltk.Frame(C_("heading", "Search"), child=vb)

            super().__init__(spacing=MARGIN)
            self.set_border_width(MARGIN)
            self.title = _("Browsers")
            self.pack_start(create_search_frame(), False, True, TOP_MARGIN)
            self.pack_start(create_display_frame(), False, True, MARGIN)

            # Ratings
            c1 = CS(
                _("Confirm _multiple ratings"),
                "browsers",
                "rating_confirm_multiple",
                populate=True,
                tooltip=_(
                    "Ask for confirmation before changing the "
                    "rating of multiple songs at once"
                ),
            )

            c2 = CS(
                _("Enable _one-click ratings"),
                "browsers",
                "rating_click",
                populate=True,
                tooltip=_(
                    "Enable rating by clicking on the rating column in the song list"
                ),
            )

            vbox = Gtk.VBox(spacing=MARGIN)
            vbox.pack_start(c1, False, True, 0)
            vbox.pack_start(c2, False, True, 0)
            f = qltk.Frame(_("Ratings"), child=vbox)
            self.pack_start(f, False, True, MARGIN)

            vb = Gtk.VBox(spacing=MARGIN)

            # Filename choice algorithm config
            sw = CS(
                _("Prefer _embedded art"),
                "albumart",
                "prefer_embedded",
                populate=True,
                tooltip=_(
                    "Choose to use artwork embedded in the audio "
                    "(where available) over other sources"
                ),
            )
            vb.pack_start(sw, False, True, 0)

            preferred_image_filename_tooltip = _(
                "The album art image file(s) to use when available "
                "(supports wildcards). If you want to supply more "
                "than one, separate them with commas."
            )

            sw = CS(
                _("_Preferred fixed image filename(s)"),
                "albumart",
                "force_filename",
                populate=True,
                tooltip=preferred_image_filename_tooltip,
            )
            vb.pack_start(sw, False, True, 0)

            entry = UndoEntry()
            entry.set_tooltip_text(preferred_image_filename_tooltip)
            entry.set_text(config.get("albumart", "filename"))
            entry.connect("changed", self.__changed_text, "filename")
            # Disable entry when not forcing
            entry.set_sensitive(sw.get_active())
            sw.connect("notify::active", self.__activated_force_filename, entry)
            self.__activated_force_filename(sw, None, entry)
            hb = Gtk.Box()
            entry.set_size_request(250, -1)
            hb.pack_start(entry, False, True, 12)
            vb.pack_start(hb, False, False, 0)

            f = qltk.Frame(_("Album Art"), child=vb)
            self.pack_start(f, False, True, MARGIN)

            for child in self.get_children():
                child.show_all()

        def __changed_text(self, entry, name):
            config.set("albumart", name, entry.get_text())

        def __activated_force_filename(self, switch, state, fn_entry):
            fn_entry.set_sensitive(switch.get_active())

        def _entry(self, entry, name, section="settings"):
            config.set(section, name, entry.get_text())

    class Player(Gtk.VBox):
        name = "playback"

        def _gain_scale_for(self, adj: Gtk.Adjustment) -> Gtk.Scale:
            def format_gain(scale, value):
                return f"{value:.0f} dB"

            scale = Gtk.Scale.new(Gtk.Orientation.HORIZONTAL, adj)
            scale.set_hexpand(True)
            scale.set_show_fill_level(True)
            scale.set_property("round-digits", 0)
            scale.set_value_pos(Gtk.PositionType.LEFT)
            scale.connect("format-value", format_gain)
            return scale

        def __init__(self):
            super().__init__(spacing=MARGIN)
            self.set_border_width(12)
            self.title = _("Playback")

            self.pack_start(self.create_behavior_frame(), False, True, TOP_MARGIN)

            # player backend
            if app.player and hasattr(app.player, "PlayerPreferences"):
                player_prefs = app.player.PlayerPreferences()
                f = qltk.Frame(_("Output Configuration"), child=player_prefs)
                self.pack_start(f, False, True, MARGIN)

            fallback_gain = config.getfloat("player", "fallback_gain", 0.0)
            adj = Gtk.Adjustment.new(fallback_gain, -12.0, 6.0, 0.5, 1, 0.0)
            adj.connect("value-changed", self.__changed, "player", "fallback_gain")
            fb_scale = self._gain_scale_for(adj)
            fb_scale.set_tooltip_text(
                _(
                    "If no Replay Gain information is available "
                    "for a song, scale the volume by this value"
                )
            )

            fb_label = Gtk.Label(label=_("_Fall-back gain:"))
            fb_label.set_use_underline(True)
            fb_label.set_alignment(0, 0.5)
            fb_label.set_mnemonic_widget(fb_scale)

            pre_amp_gain = config.getfloat("player", "pre_amp_gain", 0.0)
            adj = Gtk.Adjustment.new(pre_amp_gain, -12, 12, 0.5, 1, 0)
            adj.connect("value-changed", self.__changed, "player", "pre_amp_gain")
            pre_scale = self._gain_scale_for(adj)
            pre_scale.set_tooltip_text(
                _(
                    "Scale volume for all songs by this value, "
                    "as long as the result will not clip"
                )
            )

            pre_label = Gtk.Label(label=_("_Pre-amp gain:"))
            pre_label.set_use_underline(True)
            pre_label.set_alignment(0, 0.5)
            pre_label.set_mnemonic_widget(pre_scale)

            widgets = [pre_label, pre_scale, fb_label, fb_scale]
            enable_rg = CS(
                _("_Enable Replay Gain volume adjustment"),
                "player",
                "replaygain",
                populate=True,
            )
            enable_rg.connect("notify::active", self.__activated_gain, widgets)
            self.__activated_gain(enable_rg, None, widgets)

            # packing
            vb = Gtk.Box(orientation=Gtk.Orientation.VERTICAL, spacing=12)
            vb.pack_start(enable_rg, False, False, 0)

            grid = Gtk.Grid(column_spacing=6, row_spacing=3)
            grid.attach(fb_label, 0, 0, 1, 1)
            grid.attach(fb_scale, 1, 0, 1, 1)
            grid.attach(pre_label, 0, 1, 1, 1)
            grid.attach(pre_scale, 1, 1, 1, 1)
            hb = Gtk.Box()
            hb.pack_start(grid, True, True, 12)
            vb.pack_start(hb, False, True, 0)
            f = qltk.Frame(_("Replay Gain Volume Adjustment"), child=vb)
            self.pack_start(f, False, True, MARGIN)

            for child in self.get_children():
                child.show_all()

        def create_behavior_frame(self):
            vbox = Gtk.VBox()
            continue_play = CS(
                _("_Continue playback on startup"),
                "player",
                "restore_playing",
                populate=True,
                tooltip=_(
                    "If music is playing on shutdown, automatically "
                    "start playing on next startup"
                ),
            )
            vbox.pack_start(continue_play, False, False, 0)
            return qltk.Frame(_("Behavior"), child=vbox)

        def __activated_gain(self, activator, state, widgets):
            if app.player:
                # tests
                app.player.reset_replaygain()
            for widget in widgets:
                widget.set_sensitive(activator.get_active())

        def __changed(self, adj, section, name):
            config.set(section, name, str(adj.get_value()))
            app.player.reset_replaygain()

    class Tagging(Gtk.VBox):
        name = "tagging"

        def ratings_vbox(self):
            """Returns a new VBox containing all ratings widgets"""
            vb = Gtk.VBox(spacing=12)

            # Default Rating
            model = Gtk.ListStore(float)
            default_combo = Gtk.ComboBox(model=model)
            default_lab = Gtk.Label(label=_("_Default rating:"))
            default_lab.set_use_underline(True)
            default_lab.set_alignment(0, 0.5)

            def draw_rating(column, cell, model, it, data):
                num = model[it][0]
                text = f"{num:0.2f}: {util.format_rating(num)}"
                cell.set_property("text", text)

            def default_rating_changed(combo, model):
                it = combo.get_active_iter()
                if it is None:
                    return
                RATINGS.default = model[it][0]
                qltk.redraw_all_toplevels()

            def populate_default_rating_model(combo, num):
                model = combo.get_model()
                model.clear()
                deltas = []
                default = RATINGS.default
                precision = RATINGS.precision
                for i in range(num + 1):
                    r = i * precision
                    model.append(row=[r])
                    deltas.append((abs(default - r), i))
                active = sorted(deltas)[0][1]
                print_d(
                    "Choosing #%d (%.2f), closest to current %.2f"
                    % (active, precision * active, default)
                )
                combo.set_active(active)

            cell = Gtk.CellRendererText()
            default_combo.pack_start(cell, True)
            default_combo.set_cell_data_func(cell, draw_rating, None)
            default_combo.connect("changed", default_rating_changed, model)
            default_lab.set_mnemonic_widget(default_combo)

            def refresh_default_combo(num):
                populate_default_rating_model(default_combo, num)

            # Rating Scale
            model = Gtk.ListStore(int)
            scale_combo = Gtk.ComboBox(model=model)
            scale_lab = Gtk.Label(label=_("Rating _scale:"))
            scale_lab.set_use_underline(True)
            scale_lab.set_mnemonic_widget(scale_combo)

            cell = Gtk.CellRendererText()
            scale_combo.pack_start(cell, False)
            num = RATINGS.number
            for i in [1, 2, 3, 4, 5, 6, 8, 10]:
                it = model.append(row=[i])
                if i == num:
                    scale_combo.set_active_iter(it)

            def draw_rating_scale(column, cell, model, it, data):
                num_stars = model[it][0]
                text = "%d: %s" % (num_stars, RATINGS.full_symbol * num_stars)
                cell.set_property("text", text)

            def rating_scale_changed(combo, model):
                it = combo.get_active_iter()
                if it is None:
                    return
                RATINGS.number = num = model[it][0]
                refresh_default_combo(num)

            refresh_default_combo(RATINGS.number)
            scale_combo.set_cell_data_func(cell, draw_rating_scale, None)
            scale_combo.connect("changed", rating_scale_changed, model)

            default_align = Align(halign=Gtk.Align.START)
            default_align.add(default_lab)
            scale_align = Align(halign=Gtk.Align.START)
            scale_align.add(scale_lab)

            grid = create_grid()
            grid.add(scale_align)
            grid.add(scale_combo)
            grid.attach(default_align, 0, 1, 1, 1)
            grid.attach(default_combo, 1, 1, 1, 1)
            vb.pack_start(grid, False, False, 12)

            # Bayesian Factor
            bayesian_factor = config.getfloat("settings", "bayesian_rating_factor", 0.0)
            adj = Gtk.Adjustment.new(bayesian_factor, 0.0, 10.0, 0.5, 0.5, 0.0)
            bayes_spin = Gtk.SpinButton(adjustment=adj, numeric=True)
            bayes_spin.set_digits(1)
            bayes_spin.connect(
                "changed",
                self.__changed_and_signal_library,
                "settings",
                "bayesian_rating_factor",
            )
            bayes_spin.set_tooltip_text(
                _(
                    "Bayesian Average factor (C) for aggregated ratings.\n"
                    "0 means a conventional average, higher values mean that "
                    "albums with few tracks will have less extreme ratings. "
                    "Changing this value triggers a re-calculation for all "
                    "albums."
                )
            )
            bayes_label = Gtk.Label(label=_("_Bayesian averaging amount:"))
            bayes_label.set_use_underline(True)
            bayes_label.set_mnemonic_widget(bayes_spin)

            # Save Ratings
            vb.pack_start(hbox_for(bayes_label, bayes_spin, False), True, True, 0)
            sw = CS(
                _("Save ratings and play _counts in tags"),
                "editing",
                "save_to_songs",
                populate=True,
            )

            def update_entry(widget, state, email_entry):
                email_entry.set_sensitive(widget.get_active())

            vb.pack_start(sw, True, True, 0)
            lab = Gtk.Label(label=_("_Email:"))
            entry = UndoEntry()
            entry.set_tooltip_text(
                _(
                    "Ratings and play counts will be saved "
                    "in tags for this email address"
                )
            )
            entry.set_text(config.get("editing", "save_email"))
            entry.connect("changed", self.__changed, "editing", "save_email")

            # Disable the entry if not saving to tags
            sw.connect("notify::active", update_entry, entry)
            update_entry(sw, None, entry)

            lab.set_mnemonic_widget(entry)
            lab.set_use_underline(True)
            vb.pack_start(hbox_for(lab, entry), True, True, 0)

            return vb

        def tag_editing_vbox(self):
            """Returns a new VBox containing all tag editing widgets"""
            vbox = Gtk.VBox(spacing=12)
            sw = CS(
                _("_Auto-save tag changes"),
                "editing",
                "auto_save_changes",
                populate=True,
                tooltip=_(
                    "Save changes to tags without confirmation "
                    "when editing multiple files"
                ),
            )
            vbox.pack_start(sw, False, False, 0)

            def revert_split(entry, button, _, section, option):
                config.reset(section, option)
                entry.set_text(config.get(section, option))

            split_entry = ClearEntry()
            gicon = Gio.ThemedIcon.new_from_names(["edit-clear-symbolic", "edit-clear"])
            split_entry.set_icon_from_gicon(Gtk.EntryIconPosition.SECONDARY, gicon)
            split_entry.connect("icon-release", revert_split, "editing", "split_on")
            split_entry.set_text(config.get("editing", "split_on"))
            split_entry.connect("changed", self.__changed, "editing", "split_on")
            split_entry.set_tooltip_text(
                _(
                    "A set of separators to use when splitting tag values "
                    "in the tag editor. "
                    "The list is space-separated."
                )
            )
            split_label = Gtk.Label(label=_("Split _tag on:"))
            split_label.set_use_underline(True)
            split_label.set_mnemonic_widget(split_entry)

            vbox.pack_start(hbox_for(split_label, split_entry), False, False, 0)

            sub_entry = ClearEntry()
            sub_entry.enable_clear_button()
            sub_entry.set_text(config.get("editing", "sub_split_on"))
            sub_entry.connect("changed", self.__changed, "editing", "sub_split_on")
            sub_entry.connect("icon-release", revert_split, "editing", "sub_split_on")

            sub_entry.set_tooltip_text(
                _(
                    "A set of separators to use when extracting subtags from "
                    "tags in the tag editor. "
                    "The list is space-separated, and each entry must only "
                    "contain two characters."
                )
            )

            sub_label = Gtk.Label(label=_("Split _subtag on:"))
            sub_label.set_use_underline(True)
            sub_label.set_mnemonic_widget(split_entry)

            vbox.pack_start(hbox_for(sub_label, sub_entry), False, False, 0)
            return vbox

        def __init__(self):
            super().__init__(spacing=MARGIN)
            self.set_border_width(12)
            self.title = _("Tags")
            self._songs = []

            f = qltk.Frame(_("Tag Editing"), child=(self.tag_editing_vbox()))
            self.pack_start(f, False, True, TOP_MARGIN)

            f = qltk.Frame(_("Ratings"), child=self.ratings_vbox())
            self.pack_start(f, False, True, MARGIN)

            for child in self.get_children():
                child.show_all()

        def __changed(self, entry, section, name):
            config.set(section, name, entry.get_text())

        def __changed_and_signal_library(self, entry, section, name):
            config.set(section, name, str(entry.get_value()))
            print_d('Signalling "changed" to entire library. Hold tight...')
            # Cache over clicks
            self._songs = self._songs or list(app.library.values())
            copool.add(
                emit_signal,
                self._songs,
                funcid="library changed",
                name=_("Updating for new ratings"),
            )

    class Library(Gtk.VBox):
        name = "library"

        def __init__(self):
            super().__init__(spacing=MARGIN)
            self.set_border_width(12)
            self.title = _("Library")

            def refresh_cb(button):
                scan_library(app.library, force=False)

            refresh = qltk.Button(_("_Scan Library"), Icons.VIEW_REFRESH)
            refresh.connect("clicked", refresh_cb)
            refresh.set_tooltip_text(_("Check for changes in your library"))

            def reload_cb(button):
                scan_library(app.library, force=True)

            reload_ = qltk.Button(_("Re_build Library"), Icons.DOCUMENT_NEW)
            reload_.connect("clicked", reload_cb)
            reload_.set_tooltip_text(
                _("Reload all songs in your library. This can take a long time.")
            )

            vb = Gtk.Box(orientation=Gtk.Orientation.VERTICAL, spacing=12)
            hbox = Gtk.Box()
            hbox.pack_end(reload_, False, True, 0)
            hbox.pack_end(refresh, False, True, 12)
            vb.pack_start(hbox, False, True, 0)

            self.pack_start(self.create_behavior_frame(), False, False, TOP_MARGIN)
            self.pack_start(self.create_scandirs_frame(), False, True, MARGIN)

            # during testing
            if app.library is not None:
                masked = MaskedBox(app.library)
                f = qltk.Frame(_("Hidden Songs"), child=masked)
                self.pack_start(f, False, True, 12)

            for child in self.get_children():
                child.show_all()

        def create_behavior_frame(self):
            vb = Gtk.Box(orientation=Gtk.Orientation.VERTICAL, spacing=12)
            scan_at_start_sw = CS(
                _("Scan library _on start"),
                "library",
                "refresh_on_start",
                populate=True,
            )
            req_restart = _("A restart is required for any changes to take effect")
            watch_lib_sw = CS(
                _("_Watch directories for changes"),
                "library",
                "watch",
                populate=True,
                tooltip=_(
                    "Watch library directories for external file additions, "
                    "deletions and renames."
                )
                + "\n"
                + req_restart,
            )
            vb.pack_start(watch_lib_sw, False, True, 0)
            vb.pack_start(scan_at_start_sw, False, True, 0)
            return qltk.Frame(_("Behavior"), child=vb)

        def create_scandirs_frame(self):
            vb = Gtk.Box(orientation=Gtk.Orientation.VERTICAL, spacing=12)
            scan_dirs = ScanBox()
            vb.pack_start(scan_dirs, True, True, 0)
            return qltk.Frame(_("Scan Directories"), child=vb)

    class Advanced(Gtk.VBox):
        name = "advanced"

        def __init__(self):
            super().__init__(spacing=MARGIN)
            self.set_border_width(12)
            self.title = _("Advanced")
            scrolledwin = Gtk.ScrolledWindow()
            scrolledwin.set_policy(Gtk.PolicyType.AUTOMATIC, Gtk.PolicyType.AUTOMATIC)
            scrolledwin.add_with_viewport(AdvancedPreferencesPane())
            self.pack_start(scrolledwin, True, True, 3)

    def __init__(self, parent, open_page=None, all_pages=True):
        if self.is_not_unique():
            return
        super().__init__()
        self.current_scan_dirs = get_scan_dirs()
        self.set_title(_("Preferences"))
        self.set_resizable(True)
        self.set_transient_for(qltk.get_top_parent(parent))

        self.__notebook = notebook = qltk.Notebook()
        pages = [self.Tagging]
        if all_pages:
            pages = (
                [self.SongList, self.Browsers, self.Player, self.Library]
                + pages
                + [self.Advanced]
            )
        for Page in pages:
            page = Page()
            page.show()
            notebook.append_page(page)
        if len(pages) > 1:
            add_css(notebook, "tab { padding: 6px 24px } ")
        else:
            notebook.set_show_tabs(False)

        if open_page in [page.name for page in pages]:
            self.set_page(open_page)
        else:
            page_name = config.get("memory", "prefs_page", "")
            self.set_page(page_name)

        def on_switch_page(notebook, page, page_num):
            config.set("memory", "prefs_page", page.name)

        notebook.connect("switch-page", on_switch_page)

        close = Button(_("_Close"), Icons.WINDOW_CLOSE)
        connect_obj(close, "clicked", lambda x: x.destroy(), self)
        button_box = Gtk.HButtonBox()
        button_box.set_layout(Gtk.ButtonBoxStyle.END)
        button_box.pack_start(close, True, True, 0)

        self.use_header_bar()
        if self.has_close_button():
            self.set_border_width(0)
            notebook.set_show_border(False)
            self.add(notebook)
        else:
            self.set_border_width(12)
            vbox = Gtk.VBox(spacing=12)
            vbox.pack_start(notebook, True, True, 0)
            vbox.pack_start(button_box, False, True, 0)
            self.add(vbox)

        connect_obj(self, "destroy", PreferencesWindow.__destroy, self)

        self.get_child().show_all()

    def set_page(self, name):
        notebook = self.__notebook
        for p in range(notebook.get_n_pages()):
            if notebook.get_nth_page(p).name == name:
                notebook.set_current_page(p)

    def __destroy(self):
        config.save()
        new_dirs = set(get_scan_dirs())
        gone_dirs = set(self.current_scan_dirs) - new_dirs
        if new_dirs - set(self.current_scan_dirs):
            print_d("Library paths have been added, re-scanning...")
            scan_library(app.library, force=False)
        elif gone_dirs:
            copool.add(app.librarian.remove_roots, gone_dirs)


def create_grid(column_spacing: int = 12, row_spacing: int = 6):
    return Gtk.Grid(row_spacing=row_spacing, column_spacing=column_spacing)


def hbox_for(label: Gtk.Label, entry: Gtk.Entry, expand_entry: bool = True) -> Gtk.Box:
    hb = Gtk.Box(spacing=12)
    hb.pack_start(label, False, False, 0)
    hb.pack_end(entry, expand_entry, True, 0)
    return hb



# Copyright 2005 Joe Wreschnig, Michael Urman
#           2013 Christoph Reiter
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

from gi.repository import Gtk, Gdk, GLib

from quodlibet import _
from quodlibet import config
from quodlibet import qltk
from quodlibet import util
from quodlibet.qltk import get_top_parent
from quodlibet.qltk.tracker import TimeTracker
from quodlibet.qltk import bookmarks
from quodlibet.qltk.x import Align
from quodlibet.qltk import Icons
from quodlibet.qltk.ccb import ConfigCheckMenuItem
from quodlibet.qltk.util import (
    window_grab_and_map,
    window_ungrab_and_unmap,
    position_window_beside_widget,
)
from quodlibet.qltk.x import SeparatorMenuItem
from quodlibet.util import connect_obj, connect_destroy


class TimeLabel(Gtk.Label):
    """A label for displaying the running time

    It tries to minimize size changes due to unequal character widths
    with the same number of characters.

    e.g. a time display -> 04:20
    """

    def __init__(self, time_=0):
        Gtk.Label.__init__(self)
        self.__widths = {}  # num-chars -> (max-min-width, max-natural-width)
        self._disabled = False
        self.set_time(time_)

    def do_get_preferred_width(self):
        widths = Gtk.Label.do_get_preferred_width(self)

        # If for same number of characters, the needed width was larger,
        # use that instead of the current one
        num_chars = len(self.get_text())
        max_widths = self.__widths.get(num_chars, widths)
        widths = max(widths[0], max_widths[0]), max(widths[1], max_widths[1])
        self.__widths[num_chars] = widths
        return widths

    def set_time(self, time_):
        """Set the time in seconds"""

        self._last_time = time_
        if self._disabled:
            return
        self.set_text(util.format_time_display(time_))

    def set_disabled(self, disabled):
        """Disable the time display temporarily, means there is no meaningful
        time to show. Re-enabling will show the previous time value
        """

        self._disabled = disabled
        if disabled:
            self.set_text("‒\u2236‒‒")
        else:
            self.set_time(self._last_time)


class HSlider(Gtk.Button):
    def __init__(self, child=None):
        super().__init__()
        if child:
            self.add(child)
        self.connect("clicked", self.__clicked)

        self._disable_slider = False
        self.__grabbed = []

        window = self.__window = Gtk.Window(type=Gtk.WindowType.POPUP)
        self.__adj = Gtk.Adjustment.new(0, 0, 0, 3, 15, 0)

        frame = Gtk.Frame()
        frame.set_border_width(0)
        frame.set_shadow_type(Gtk.ShadowType.OUT)

        self.add_events(Gdk.EventMask.SCROLL_MASK)

        hscale = Gtk.Scale(adjustment=self.__adj)
        hscale.set_orientation(Gtk.Orientation.HORIZONTAL)
        window.connect("button-press-event", self.__button)
        window.connect("key-press-event", self.__key)
        hscale.set_draw_value(False)
        self.scale = hscale
        window.add(frame)
        self._box = Gtk.Box(orientation=Gtk.Orientation.HORIZONTAL)
        self._box.add(hscale)
        frame.add(self._box)
        self.connect("scroll-event", self.__scroll, hscale)

        self.connect("destroy", self.__destroy)

        # forward scroll event to the button
        def foward_scroll(scale, event):
            self.emit("scroll-event", event.copy())

        window.connect("scroll-event", foward_scroll)

        # ignore scroll events on the scale, the window handles it instead
        self.scale.connect("scroll-event", lambda *x: True)

        # handle all unhandled button events on the scale
        # so only events not on the scale hide the window
        def handle_all(scale, event):
            return True

        self.scale.connect_after("button-press-event", handle_all)
        self.scale.connect_after("button-release-event", handle_all)

        # forward release event to the scale
        def foward_release(scale, event):
            self.scale.emit("button-release-event", event.copy())

        window.connect("button-release-event", foward_release)

        self.set_slider_length(200)

        if child:
            self.get_child().show_all()

    def __destroy(self, *args):
        self.__window.destroy()
        self.__window = None

    def set_slider_disabled(self, disable):
        """Hide the slider and don't allow showing it again until it is
        enabled again
        """

        self._disable_slider = disable
        if disable:
            self.__popup_hide()

    def set_slider_length(self, length):
        self.scale.set_size_request(length, -1)

        # force a window resize..
        self.__window.resize(1, 1)

    def set_slider_widget(self, widget):
        self._box.pack_start(Align(widget, border=6, left=-3), False, True, 0)

    def __clicked(self, button):
        if self.__window.get_property("visible"):
            return

        if self._disable_slider:
            return

        if self.__grabbed:
            self.__popup_hide()

        window = self.__window
        frame = window.get_child()
        frame.show_all()

        window.set_transient_for(get_top_parent(self))
        # this type hint tells the wayland backend to create a popup
        window.set_type_hint(Gdk.WindowTypeHint.DROPDOWN_MENU)

        position_window_beside_widget(window, self)

        self.__grabbed = window_grab_and_map(
            window,
            Gdk.EventMask.BUTTON_PRESS_MASK
            | Gdk.EventMask.BUTTON_RELEASE_MASK
            | Gdk.EventMask.BUTTON_MOTION_MASK
            | Gdk.EventMask.POINTER_MOTION_MASK
            | Gdk.EventMask.SCROLL_MASK,
        )

    def __scroll(self, widget, event, hscale):
        adj = self.__adj
        v = hscale.get_value()
        direction = event.direction
        if direction in [Gdk.ScrollDirection.DOWN, Gdk.ScrollDirection.RIGHT]:
            v -= adj.props.step_increment
        elif direction in [Gdk.ScrollDirection.UP, Gdk.ScrollDirection.LEFT]:
            v += adj.props.step_increment
        else:
            # newer Gdk.ScrollDirection.SMOOTH
            return
        v = min(adj.props.upper, max(adj.props.lower, v))
        hscale.set_value(v)

    def __button(self, widget, ev):
        self.__popup_hide()

    def __key(self, hscale, ev):
        if ev.string in ["\n", "\r", " ", "\x1b"]:  # enter, space, escape
            self.__popup_hide()

    def __popup_hide(self):
        window_ungrab_and_unmap(self.__window, self.__grabbed)
        del self.__grabbed[:]


class SeekButton(HSlider):
    __lock = False
    __sig = None
    __seekable = True

    def __init__(self, player, library):
        hbox = Gtk.HBox(spacing=3)
        l = TimeLabel()
        self._time_label = l
        hbox.pack_start(l, True, True, 0)
        arrow = Gtk.Arrow.new(Gtk.ArrowType.RIGHT, Gtk.ShadowType.NONE)
        hbox.pack_start(arrow, False, True, 0)
        super().__init__(hbox)

        self._slider_label = TimeLabel()
        self.set_slider_widget(self._slider_label)

        self._on_seekable_changed(player)
        connect_destroy(player, "notify::seekable", self._on_seekable_changed)

        self.scale.connect("button-press-event", self.__seek_lock)
        self.scale.connect("button-release-event", self.__seek_unlock, player)
        self.scale.connect("key-press-event", self.__seek_lock)
        self.scale.connect("key-release-event", self.__seek_unlock, player)
        self.connect("scroll-event", self.__scroll, player)
        self.scale.connect("value-changed", self.__update_time, l)

        m = Gtk.Menu()
        c = ConfigCheckMenuItem(_("Display remaining time"), "player", "time_remaining")
        c.set_active(config.getboolean("player", "time_remaining"))
        connect_obj(c, "toggled", self.scale.emit, "value-changed")
        self.__remaining = c
        m.append(c)
        m.append(SeparatorMenuItem())
        i = qltk.MenuItem(_("_Edit Bookmarks…"), Icons.EDIT)

        def edit_bookmarks_cb(menu_item):
            window = bookmarks.EditBookmarks(self, library, player)
            window.show()

        i.connect("activate", edit_bookmarks_cb)
        m.append(i)
        m.show_all()

        connect_obj(self, "button-press-event", self.__check_menu, m, player, c)
        connect_obj(self, "popup-menu", self.__popup_menu, m, player)

        timer = TimeTracker(player)
        connect_obj(timer, "tick", self.__check_time, player)

        connect_destroy(library, "changed", self.__songs_changed, player, m)

        connect_destroy(player, "song-started", self.__song_started, m)
        connect_destroy(player, "seek", self.__seeked)

    def _on_seekable_changed(self, player, *args):
        self._time_label.set_disabled(not player.seekable)
        self.set_slider_disabled(not player.seekable)

    def __check_menu(self, menu, event, player, remaining_item):
        if event.type != Gdk.EventType.BUTTON_PRESS:
            return None

        if event.button == Gdk.BUTTON_SECONDARY:
            return self.__popup_menu(menu, player, event)
        if event.button == Gdk.BUTTON_MIDDLE:
            remaining_item.set_active(not remaining_item.get_active())
            return True
        return None

    def __popup_menu(self, menu, player, event=None):
        for child in menu.get_children()[2:-1]:
            menu.remove(child)
            child.destroy()

        try:
            marks = player.song.bookmarks
        except AttributeError:
            # song is None
            pass
        else:
            items = qltk.bookmarks.MenuItems(marks, player, self.__seekable)
            items.reverse()
            for i in items:
                menu.insert(i, 2)

        if event:
            qltk.popup_menu_at_widget(menu, self, 3, event.time)
        else:
            time = Gtk.get_current_event_time()
            qltk.popup_menu_under_widget(menu, self, 3, time)
        return True

    def __seeked(self, player, song, ms):
        self.scale.set_value(ms / 1000.0)

    def __scroll(self, widget, event, player):
        self.__lock = True
        if self.__sig is not None:
            GLib.source_remove(self.__sig)
        self.__sig = GLib.timeout_add(100, self.__scroll_timeout, player)

    def __scroll_timeout(self, player):
        self.__lock = False
        if self.__seekable:
            player.seek(self.scale.get_value() * 1000)
        self.__sig = None

    def __seek_lock(self, scale, event):
        self.__lock = True

    def __seek_unlock(self, scale, event, player):
        self.__lock = False
        if self.__seekable:
            player.seek(self.scale.get_value() * 1000)

    def __check_time(self, player):
        # When the song is paused GStreamer returns < 1 for position
        # queries, so if it's paused just ignore it.
        if not (player.paused or self.__lock):
            position = player.get_position() / 1000.0
            if (
                not self.__seekable
                and position > self.scale.get_adjustment().get_upper()
            ):
                self.scale.set_range(0, position)
            self.scale.set_value(position)
        return True

    def __update_time(self, scale, timer):
        value = scale.get_value()
        max_ = scale.get_adjustment().get_upper()
        remaining = value - max_
        if self.__remaining.get_active():
            remaining, value = value, remaining
        timer.set_time(value)
        self._slider_label.set_time(remaining)

    def __songs_changed(self, library, songs, player, menu):
        song = player.song
        if song in songs:
            self.__update_slider(song, menu)

    def __song_started(self, player, song, menu):
        self.scale.set_value(0)
        self.__update_slider(song, menu)

    def __update_slider(self, song, menu):
        if song and song("~#length") > 0:
            self.scale.set_range(0, song("~#length"))
            slider_width = int(song("~#length") / 1.5) + 80
            self.__seekable = True
        else:
            self.scale.set_range(0, 1)
            slider_width = 0
            self.__seekable = False

        slider_width = min(max(slider_width, 170), 400)
        self.set_slider_length(slider_width)

        for child in menu.get_children()[2:-1]:
            menu.remove(child)
            child.destroy()
        menu.get_children()[-1].set_sensitive(self.__seekable)
        self.scale.emit("value-changed")



# Copyright 2016-23 Nick Boultbee
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

import json
from datetime import datetime
from typing import Any
from urllib.parse import urlencode

from gi.repository import GObject, Gio, Soup, GLib

from quodlibet import util, config
from quodlibet.formats import AudioFile
from quodlibet.util import website
from quodlibet.util.dprint import print_w, print_d
from quodlibet.util.http import download_json, download, HTTPRequest, FailureCallback
from .library import SoundcloudFile
from .util import json_callback, Wrapper, sanitise_tag, DEFAULT_BITRATE, EPOCH


class RestApi(GObject.Object):
    """Semi-generic REST API client, using libsoup / `http.py`"""

    def __init__(self, root, on_failure: FailureCallback):
        super().__init__()
        self._cancellable = Gio.Cancellable.new()
        self.root = root
        self._on_failure = on_failure
        self.access_token = None

    def _default_params(self):
        return {}

    def _get(self, path, callback, data=None, **kwargs):
        args = self._default_params()
        args.update(kwargs)
        msg = self._add_auth_to(Soup.Message.new("GET", self._url(path, args)))
        download_json(msg, self._cancellable, callback, data, self._on_failure)

    def _add_auth_to(self, msg: Soup.Message) -> Soup.Message:
        if self.access_token:
            msg.get_request_headers().append(
                "Authorization", f"OAuth {self.access_token}"
            )
        return msg

    def _post(self, path, callback, **kwargs):
        args = self._default_params()
        args.update(kwargs)
        msg = self._add_auth_to(Soup.Message.new("POST", self._url(path)))
        post_body = urlencode(args)
        if not isinstance(post_body, bytes):
            post_body = post_body.encode("ascii")
        msg.set_request_body_from_bytes(
            "application/x-www-form-urlencoded", GLib.Bytes.new(post_body)
        )
        download_json(msg, self._cancellable, callback, None, self._on_failure)

    def _delete(self, path, callback, **kwargs):
        args = self._default_params()
        args.update(kwargs)
        # Turns out the SC API doesn't mind body arguments for DELETEs,
        # and as it's neater and slightly more secure, let's do that.
        body = urlencode(args)
        if not isinstance(body, bytes):
            body = body.encode("ascii")
        msg = self._add_auth_to(Soup.Message.new("DELETE", self._url(path)))
        msg.set_request_body_from_bytes(
            "application/x-www-form-urlencoded", GLib.Bytes.new(body)
        )
        download(msg, self._cancellable, callback, None, try_decode=True)

    def _url(self, path, args=None):
        path = f"{self.root}{path}"
        return f"{path}?{urlencode(args)}" if args else path


class SoundcloudApiClient(RestApi):
    __CLIENT_SECRET = "ca2b69301bd1f73985a9b47224a2a239"
    __CLIENT_ID = "5acc74891941cfc73ec8ee2504be6617"
    API_ROOT = "https://api.soundcloud.com"
    REDIRECT_URI = "https://quodlibet.github.io/callbacks/soundcloud.html"
    PAGE_SIZE = 100
    MIN_DURATION_SECS = 120
    COUNT_TAGS = {
        f"{t}_count"
        for t in (
            "playback",
            "download",
            "likes",
            "favoritings",
            "download",
            "comments",
        )
    }

    __gsignals__ = {
        "fetch-success": (GObject.SignalFlags.RUN_LAST, None, (object,)),
        "fetch-failure": (GObject.SignalFlags.RUN_LAST, None, (object,)),
        "songs-received": (GObject.SignalFlags.RUN_LAST, None, (object,)),
        "stream-uri-received": (GObject.SignalFlags.RUN_LAST, None, (object, str)),
        "comments-received": (
            GObject.SignalFlags.RUN_LAST,
            None,
            (
                int,
                object,
            ),
        ),
        "authenticated": (GObject.SignalFlags.RUN_LAST, None, (object,)),
    }

    def __init__(self):
        print_d("Starting Soundcloud API...")
        super().__init__(self.API_ROOT, self._on_failure)
        self.access_token = config.get("browsers", "soundcloud_token", None)
        self.refresh_token = config.get("browsers", "soundcloud_refresh_token", None)
        self.user_id = config.get("browsers", "soundcloud_user_id", None)
        if not self.user_id:
            self._get_me()
        self.username = None

    @property
    def online(self):
        return bool(self.access_token)

    def _on_failure(self, req: HTTPRequest, _exc: Exception, data: Any) -> None:
        """Callback for HTTP failures."""
        code = req.message.get_property("status-code")
        if code in (401,):
            print_w("User session no longer valid, logging out.")
            if self.access_token:
                # Could call log_out to persist, but we're probably about to refresh...
                self.access_token = None
                self._refresh_tokens()
            else:
                print_w("Refreshing didn't work either, oh dear.")
                self.log_out()

    def _default_params(self):
        return {}

    def authenticate_user(self):
        # create client object with app credentials
        if self.access_token:
            print_d("Ignoring saved Soundcloud token...")
        # redirect user to authorize URL
        website(self._authorize_url)

    def log_out(self):
        print_d("Destroying access token...")
        self.access_token = None
        self.user_id = None
        self.save_auth()

    def get_tokens(self, code):
        print_d("Getting access token...")
        options = {
            "grant_type": "authorization_code",
            "code": code,
            "redirect_uri": self.REDIRECT_URI,
            "client_id": self.__CLIENT_ID,
            "client_secret": self.__CLIENT_SECRET,
        }
        self._post("/oauth2/token", self._receive_tokens, **options)

    def _refresh_tokens(self):
        print_d("Refreshing access token...")
        options = {
            "grant_type": "refresh_token",
            "refresh_token": self.refresh_token,
            "client_id": self.__CLIENT_ID,
            "client_secret": self.__CLIENT_SECRET,
        }
        self._post("/oauth2/token", self._receive_tokens, **options)

    @json_callback
    def _receive_tokens(self, json, _data):
        self.access_token = json["access_token"]
        refresh_token = json.get("refresh_token", None)
        if refresh_token:
            # Just in case we don't get it...
            self.refresh_token = refresh_token
            print_d("Got refresh token.")

        print_d(f"Got an access token: ...{self.access_token[-6:]}")
        self.save_auth()
        if not self.user_id:
            self._get_me()

    def _get_me(self):
        self._get("/me", self._receive_me)

    @json_callback
    def _receive_me(self, json, _data):
        self.username = json["username"]
        self.user_id = json["id"]
        self.emit("authenticated", Wrapper(json))

    def get_tracks(self, params):
        merged = {
            "q": "",
            "limit": self.PAGE_SIZE,
            "duration[from]": self.MIN_DURATION_SECS * 1000,
            "access": "playable",
        }
        for k, v in params.items():
            delim = " " if k == "q" else ","
            merged[k] = delim.join(list(v))
        print_d(f"Getting tracks: params={merged}")
        self._get("/tracks", self._on_track_data, **merged)

    def get_stream_url(self, song):
        try:
            self._get(
                f"/tracks/{song['soundcloud_track_id']}/streams",
                self._on_track_stream_urls_data,
                song,
            )
        except Exception as e:
            print_w(f"Problem getting stream URL for {song} ({e})")

    @json_callback
    def _on_track_stream_urls_data(self, json, song):
        uri = json["http_mp3_128_url"]
        self.emit("stream-uri-received", song, uri)

    @json_callback
    def _on_track_data(self, json, _data):
        songs = list(filter(None, [self._audiofile_for(r) for r in json]))
        self.emit("songs-received", songs)

    def get_favorites(self):
        self._get("/me/likes/tracks", self._on_track_data, limit=self.PAGE_SIZE)

    def get_my_tracks(self):
        self._get("/me/tracks", self._on_track_data, limit=self.PAGE_SIZE)

    def get_comments(self, track_id):
        self._get(f"/tracks/{track_id}/comments", self._receive_comments, limit=500)

    @json_callback
    def _receive_comments(self, json, _data):
        print_d(f"Got comments: {str(json)[:255]}...")
        if json and len(json):
            # Should all be the same track...
            track_id = json[0]["track_id"]
            self.emit("comments-received", track_id, json)

    def save_auth(self):
        config.set("browsers", "soundcloud_token", self.access_token or "")
        config.set("browsers", "soundcloud_refresh_token", self.refresh_token or "")
        config.set("browsers", "soundcloud_user_id", self.user_id or "")

    def save_favorite(self, track_id):
        print_d(f"Saving track {track_id} as favorite")
        url = f"/likes/tracks/{track_id}"
        self._post(url, self._on_favorited)

    def remove_favorite(self, track_id):
        print_d(f"Deleting favorite for {track_id}")
        url = f"/likes/tracks/{track_id}"
        self._delete(url, self._on_favorited)

    @json_callback
    def _on_favorited(self, json, _data):
        print_d("Successfully updated favorite")

    def _audiofile_for(self, response) -> AudioFile | None:
        r = Wrapper(response)
        d = r.data
        try:
            # It's not a _play_ URI, because that needs a separate request,
            # but it's unique and permanent, so that's good enough here
            url = r.uri
            if not url:
                print_d(f"Unusable response (no URI): {d}")
                return None
            song = SoundcloudFile(
                uri=url,
                track_id=r.id,
                client=self,
                favorite=d.get("user_favorite", False),
            )
        except Exception as e:
            print_w(f"Track {r.id} no good ({e})")
            return None

        def get_utc_date(s):
            parts = s.split()
            dt = datetime.strptime(" ".join(parts[:-1]), "%Y/%m/%d %H:%M:%S")
            return int((dt - EPOCH).total_seconds())

        def put_time(tag, r, attr):
            try:
                song[tag] = get_utc_date(r[attr])
            except KeyError:
                pass

        def put_date(tag, r, attr):
            try:
                parts = r[attr].split()
                dt = datetime.strptime(" ".join(parts[:-1]), "%Y/%m/%d %H:%M:%S")
                song[tag] = dt.strftime("%Y-%m-%d")
            except KeyError:
                pass

        def put_counts(tags):
            for tag in tags:
                try:
                    song[f"~#{tag}"] = int(r[tag])
                except (KeyError, TypeError):
                    # Nothing we can do really.
                    pass

        try:
            song.update(
                title=r.title,
                artist=r.user["username"],
                soundcloud_user_id=str(r.user.id),
                website=r.permalink_url,
                genre="\n".join(r.genre and r.genre.split(",") or []),
            )
            song["~#bitrate"] = DEFAULT_BITRATE
            if r.description:
                song["comment"] = sanitise_tag(r.description)
            song["~#length"] = int(r.duration) / 1000
            art_url = r.artwork_url
            if art_url:
                song["artwork_url"] = art_url.replace("-large.", "-t500x500.")
            put_time("~#mtime", r, "last_modified")
            put_date("date", r, "created_at")
            put_counts(self.COUNT_TAGS)
            plays = d.get("user_playback_count", 0)
            if plays:
                song["~#playcount"] = plays
        except Exception as e:
            print_w(f"Couldn't parse song ({e!r}): {json.dumps(r._raw)}")
        return song

    @util.cached_property
    def _authorize_url(self):
        url = f"{self.API_ROOT}/connect"
        options = {
            "scope": "",
            "client_id": self.__CLIENT_ID,
            "response_type": "code",
            "redirect_uri": self.REDIRECT_URI,
        }
        return f"{url}?{urlencode(options)}"



# Copyright (C) 2012-13 Nick Boultbee, Thomas Vogt
# Copyright (C) 2008 Andreas Bombe
# Copyright (C) 2005  Michael Urman
# Based on osd.py (C) 2005 Ton van den Heuvel, Joe Wreshnig
#                 (C) 2004 Gustavo J. A. M. Carneiro
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

from quodlibet.plugins import (
    PluginConfig,
    ConfProp,
    IntConfProp,
    FloatConfProp,
    ColorConfProp,
)


DEFAULT_PATTERN = (
    r"<album|[b]<album>[/b]<discnumber| - Disc "
    """<discnumber>><part| - [b]<part>[/b]><tracknumber| - <tracknumber>>
    >[span weight='bold' size='large']<title>[/span] - <~length><version|
    [small][i]<version>[/i][/small]><~people|
    by <~people>>"""
)


def get_config(prefix):
    class AnimOsdConfig:
        plugin_conf = PluginConfig(prefix)

        font = ConfProp(plugin_conf, "font", "Sans 22")
        string = ConfProp(plugin_conf, "string", DEFAULT_PATTERN)
        pos_x = FloatConfProp(plugin_conf, "pos_x", 0.5)
        pos_y = FloatConfProp(plugin_conf, "pos_y", 0.0)
        corners = IntConfProp(plugin_conf, "corners", 1)
        delay = IntConfProp(plugin_conf, "delay", 2500)
        monitor = IntConfProp(plugin_conf, "monitor", 0)
        align = IntConfProp(plugin_conf, "align", 1)
        coversize = IntConfProp(plugin_conf, "coversize", 120)
        text = ColorConfProp(plugin_conf, "text", (0.9, 0.9, 0.9, 0.0))
        outline = ColorConfProp(plugin_conf, "outline", (-1.0, 0.0, 0.0, 0.2))
        shadow = ColorConfProp(plugin_conf, "shadow", (-1.0, 0.0, 0.0, 0.1))
        fill = ColorConfProp(plugin_conf, "fill", (0.25, 0.25, 0.25, 0.5))

    return AnimOsdConfig()



# Copyright 2005 Michael Urman
#        2016-22 Nick Boultbee
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

from quodlibet import _, ngettext
from quodlibet.qltk.notif import Task
from quodlibet.util import copool
from quodlibet.util.dprint import print_w
from quodlibet.formats import AudioFileError
from quodlibet import util
from quodlibet import qltk
from quodlibet.util.misc import total_ordering, hashable


@hashable
@total_ordering
class SongWrapper:
    __slots__ = ["_song", "_updated", "_needs_write"]

    def __init__(self, song):
        self._song = song
        self._updated = False
        self._needs_write = False

    def _was_updated(self):
        return self._updated

    def __setitem__(self, key, value):
        if key in self and self[key] == value:
            return None
        self._updated = True
        self._needs_write = self._needs_write or not key.startswith("~")
        return self._song.__setitem__(key, value)

    def __delitem__(self, key):
        retval = self._song.__delitem__(key)
        self._updated = True
        self._needs_write = self._needs_write or not key.startswith("~")
        return retval

    def __getattr__(self, attr):
        return getattr(self._song, attr)

    def __setattr__(self, attr, value):
        # Don't set our attributes on the song. However, we only want to
        # set attributes the song already has. So, if the attribute
        # isn't one of ours, and isn't one of the song's, hand it off
        # to our parent's attribute handler for error handling.
        if attr in self.__slots__:
            return super().__setattr__(attr, value)
        if hasattr(self._song, attr):
            return setattr(self._song, attr, value)
        return super().__setattr__(attr, value)

    def __hash__(self):
        return hash(self._song)

    def __eq__(self, other):
        if hasattr(other, "_song"):
            other = other._song
        return self._song == other

    def __lt__(self, other):
        if hasattr(other, "_song"):
            other = other._song
        return self._song < other

    def __getitem__(self, *args):
        return self._song.__getitem__(*args)

    def __contains__(self, key):
        return key in self._song

    def __call__(self, *args):
        return self._song(*args)

    def pop(self, *args):
        self._updated = True
        self._needs_write = True
        return self._song.pop(*args)

    def update(self, other):
        self._updated = True
        self._needs_write = True
        return self._song.update(other)

    def rename(self, newname):
        self._updated = True
        return self._song.rename(newname)


def list_wrapper(songs):
    def wrap(song):
        if song is None:
            return None
        return SongWrapper(song)

    return [wrap(s) for s in songs]


def check_wrapper_changed(library, songs):
    need_write = [s for s in songs if s._needs_write]
    if need_write:
        total = len(need_write)

        def generate():
            msg = ngettext("Saving %d file", "Saving %d files", total) % total
            with Task(_("Auto-Saving"), msg) as task:
                yield
                for i, song in enumerate(need_write):
                    try:
                        song._song.write()
                    except AudioFileError as e:
                        dialog = qltk.ErrorMessage(
                            None,
                            _("Unable to edit song"),
                            _(
                                "Saving %s failed. "
                                "The file may be read-only, corrupted, or you "
                                "do not have permission to edit it."
                            )
                            % util.bold(song("~basename")),
                            escape_desc=False,
                        )
                        dialog.run()
                        print_w(f"Couldn't save song {song('~filename')} ({e!r})")
                    else:
                        task.update((i + 1) / total)
                        yield

        # Small enough to see if there's only one or two, very small if lots
        interval = max(250 // total, 5)
        copool.add(generate, timeout=interval, funcid="update_wrapped_songs")
    _inform_library_of_changed(library, songs)


def background_check_wrapper_changed(library, songs):
    for song in songs:
        if not song._needs_write:
            continue

        try:
            song._song.write()
        except AudioFileError as e:
            print_w("Couldn't save song {} ({})".format(song("~filename"), e))

    _inform_library_of_changed(library, songs)


def _inform_library_of_changed(library, songs):
    changed = []
    for song in songs:
        if song._was_updated():
            changed.append(song._song)
        elif not song.valid() and song.exists():
            library.reload(song._song)
    library.changed(changed)



# Copyright 2004-2007 Joe Wreschnig, Michael Urman, Iñigo Serna
#           2009-2010 Steven Robertson
#           2012-2023 Nick Boultbee
#           2009-2014 Christoph Reiter
#           2022 Thomas Leberbauer
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.


import os

from gi.repository import GLib, Gtk, Gdk, Gio

from .prefs import Preferences, DEFAULT_PATTERN_TEXT

import quodlibet
from quodlibet import app
from quodlibet import ngettext
from quodlibet import config
from quodlibet import qltk
from quodlibet import util
from quodlibet import _
from quodlibet.browsers import Browser
from quodlibet.browsers.albums.main import (
    AlbumTagCompletion,
    PreferencesButton as AlbumPreferencesButton,
)
from quodlibet.browsers.covergrid.models import (
    AlbumListFilterModel,
    AlbumListModel,
    AlbumListSortModel,
)
from quodlibet.browsers.covergrid.widgets import AlbumWidget
from quodlibet.browsers._base import DisplayPatternMixin
from quodlibet.query import Query
from quodlibet.qltk.information import Information
from quodlibet.qltk.properties import SongProperties
from quodlibet.qltk.songsmenu import SongsMenu
from quodlibet.qltk.x import MenuItem, Align, ScrolledWindow, RadioMenuItem
from quodlibet.qltk.x import SymbolicIconImage
from quodlibet.qltk.searchbar import SearchBarBox
from quodlibet.qltk.menubutton import MenuButton
from quodlibet.qltk import Icons
from quodlibet.util import connect_destroy
from quodlibet.util import connect_obj
from quodlibet.qltk import popup_menu_at_widget


class PreferencesButton(AlbumPreferencesButton):
    def __init__(self, browser, model):
        Gtk.HBox.__init__(self)

        sort_orders = [
            (_("_Title"), self.__compare_title),
            (_("_People"), self.__compare_people),
            (_("_Date"), self.__compare_date),
            (_("_Date Added"), self.__compare_date_added),
            (_("_Original Date"), self.__compare_original_date),
            (_("_Genre"), self.__compare_genre),
            (_("_Rating"), self.__compare_rating),
            (_("Play_count"), self.__compare_avgplaycount),
        ]

        menu = Gtk.Menu()

        sort_item = Gtk.MenuItem(label=_("Sort _by…"), use_underline=True)
        sort_menu = Gtk.Menu()

        active = config.getint("browsers", "album_sort", 1)

        item = None
        for i, (label, func) in enumerate(sort_orders):
            item = RadioMenuItem(group=item, label=label, use_underline=True)
            model.set_sort_func(100 + i, func)
            if i == active:
                model.set_sort_column_id(100 + i, Gtk.SortType.ASCENDING)
                item.set_active(True)
            item.connect(
                "toggled", util.DeferredSignal(self.__sort_toggled_cb), model, i
            )
            sort_menu.append(item)

        sort_item.set_submenu(sort_menu)
        menu.append(sort_item)

        pref_item = MenuItem(_("_Preferences"), Icons.PREFERENCES_SYSTEM)
        menu.append(pref_item)
        connect_obj(pref_item, "activate", Preferences, browser)

        menu.show_all()

        button = MenuButton(
            SymbolicIconImage(Icons.EMBLEM_SYSTEM, Gtk.IconSize.MENU), arrow=True
        )
        button.set_menu(menu)
        self.pack_start(button, True, True, 0)


class CoverGridContainer(ScrolledWindow):
    def __init__(self, fb):
        super().__init__(
            hscrollbar_policy=Gtk.PolicyType.NEVER,
            vscrollbar_policy=Gtk.PolicyType.AUTOMATIC,
            shadow_type=Gtk.ShadowType.IN,
        )
        self._fb = fb
        fb.set_hadjustment(self.props.hadjustment)
        fb.set_vadjustment(self.props.vadjustment)
        self.add(fb)

    def scroll_up(self):
        va = self.props.vadjustment
        va.props.value = va.props.lower

    def scroll_to_child(self, child):
        def scroll():
            va = self.props.vadjustment
            if va is None:
                return
            v = va.props.value
            coords = child.translate_coordinates(self, 0, v)
            if coords is None:
                return
            x, y = coords
            h = child.get_allocation().height
            p = va.props.page_size
            if y < v:
                va.props.value = y
            elif y + h > v + p:
                va.props.value = y - p + h

        GLib.idle_add(scroll, priority=GLib.PRIORITY_LOW)

    def do_focus(self, direction):
        is_tab = (
            direction == Gtk.DirectionType.TAB_FORWARD
            or direction == Gtk.DirectionType.TAB_BACKWARD
        )
        if not is_tab:
            self._fb.child_focus(direction)
            return True

        if self.get_focus_child():
            # [Tab] moves focus beyond this container
            return False

        children = self._fb.get_selected_children()
        if children:
            children[0].grab_focus()
        else:
            self._fb.child_focus(direction)
        return True


def _get_cover_size():
    mag = config.getfloat("browsers", "covergrid_magnification", 3.0)
    size = config.getint("browsers", "cover_size")
    if size <= 0:
        size = 48
    return mag * size


class CoverGrid(Browser, util.InstanceTracker, DisplayPatternMixin):
    __model = None

    _PATTERN_FN = os.path.join(quodlibet.get_user_dir(), "album_pattern")
    _DEFAULT_PATTERN_TEXT = DEFAULT_PATTERN_TEXT

    name = _("Cover Grid")
    accelerated_name = _("_Cover Grid")
    keys = ["CoverGrid"]
    priority = 5

    def pack(self, songpane):
        container = self.songcontainer
        container.pack1(self, True, False)
        container.pack2(songpane, True, False)
        return container

    def unpack(self, container, songpane):
        container.remove(songpane)
        container.remove(self)

    @classmethod
    def init(cls, library):
        super().load_pattern()

    @classmethod
    def _init_model(cls, library):
        if cls.__model is None:
            cls.__model = AlbumListModel(library)
            cls.__library = library

    @classmethod
    def _destroy_model(cls):
        cls.__model.destroy()
        cls.__model = None

    @classmethod
    def toggle_text(cls):
        text_visible = config.getboolean("browsers", "album_text", True)
        for covergrid in cls.instances():
            for child in covergrid.view:
                child.props.text_visible = text_visible

    @classmethod
    def toggle_item_all(cls):
        show = config.getboolean("browsers", "covergrid_all", True)
        for covergrid in cls.instances():
            covergrid.__model_filter.props.include_item_all = show

    @classmethod
    def toggle_wide(cls):
        wide = config.getboolean("browsers", "covergrid_wide", False)
        for covergrid in cls.instances():
            covergrid.songcontainer.set_orientation(
                Gtk.Orientation.HORIZONTAL if wide else Gtk.Orientation.VERTICAL
            )

    @classmethod
    def update_mag(cls):
        cover_size = _get_cover_size()
        for covergrid in cls.instances():
            for child in covergrid.view:
                child.cover_size = cover_size
            covergrid.view.queue_resize()

    def __init__(self, library):
        super().__init__(orientation=Gtk.Orientation.VERTICAL, spacing=6)

        self.songcontainer = qltk.paned.ConfigRVPaned("browsers", "covergrid_pos", 0.4)
        if config.getboolean("browsers", "covergrid_wide", False):
            self.songcontainer.set_orientation(Gtk.Orientation.HORIZONTAL)

        self._register_instance()
        self._init_model(library)

        self.__cover_cancel = Gio.Cancellable()

        model_sort = AlbumListSortModel(model=self.__model)
        self.__model_filter = model_filter = AlbumListFilterModel(
            include_item_all=config.getboolean("browsers", "covergrid_all", True),
            child_model=model_sort,
        )

        def create_album_widget(model):
            item_padding = config.getint("browsers", "item_padding", 6)
            text_visible = config.getboolean("browsers", "album_text", True)
            cover_size = _get_cover_size()
            widget = AlbumWidget(
                model,
                display_pattern=self.display_pattern,
                cover_size=cover_size,
                padding=item_padding,
                text_visible=text_visible,
                cancelable=self.__cover_cancel,
            )
            widget.connect("songs-menu", self.__popup)
            return widget

        self.view = view = Gtk.FlowBox(
            valign=Gtk.Align.START,
            activate_on_single_click=False,
            selection_mode=Gtk.SelectionMode.MULTIPLE,
            homogeneous=True,
            min_children_per_line=1,
            max_children_per_line=10,
            row_spacing=config.getint("browsers", "row_spacing", 6),
            column_spacing=config.getint("browsers", "column_spacing", 6),
        )

        self.scrollwin = sw = CoverGridContainer(view)

        view.connect(
            "selected-children-changed",
            util.DeferredSignal(
                lambda _: self.__update_songs(select_default=False), owner=self
            ),
        )

        targets = [
            ("text/x-quodlibet-songs", Gtk.TargetFlags.SAME_APP, 1),
            ("text/uri-list", 0, 2),
        ]
        targets = [Gtk.TargetEntry.new(*t) for t in targets]
        view.drag_source_set(
            Gdk.ModifierType.BUTTON1_MASK, targets, Gdk.DragAction.COPY
        )

        view.connect("drag-data-get", self.__drag_data_get)
        view.connect("child-activated", self.__child_activated)

        self.accelerators = Gtk.AccelGroup()
        search = SearchBarBox(
            completion=AlbumTagCompletion(), accel_group=self.accelerators
        )
        search.connect("query-changed", lambda *a: self.__update_filter())
        connect_obj(search, "focus-out", lambda w: w.grab_focus(), view)
        self.__search = search

        prefs = PreferencesButton(self, model_sort)
        search.pack_start(prefs, False, True, 0)
        self.pack_start(Align(search, left=6, top=0), False, True, 0)
        self.pack_start(sw, True, True, 0)

        self.__update_filter()
        model_filter.connect(
            "notify::filter",
            util.DeferredSignal(lambda *a: self.__update_songs(), owner=self),
        )

        self.connect("key-press-event", self.__key_pressed, library.librarian)
        self.connect("destroy", self.__destroy)

        if app.cover_manager:
            connect_destroy(app.cover_manager, "cover-changed", self.__cover_changed)

        # show all before binding the model, so a label in a flowbox child will
        # stay hidden if so configured by the "browsers.album_text" property.
        self.show_all()
        view.bind_model(model_filter, create_album_widget)

    def __update_songs(self, select_default=True):
        songs = self.__get_selected_songs(sort=False)
        if not select_default or songs:
            self.songs_selected(songs)
        else:
            child = self.view.get_child_at_index(0)
            if child:
                self.view.select_child(child)
            else:
                self.songs_selected(songs)

    def __key_pressed(self, widget, event, librarian):
        if qltk.is_accel(event, "<Primary>I"):
            songs = self.__get_selected_songs()
            if songs:
                window = Information(librarian, songs, self)
                window.show()
            return True
        if qltk.is_accel(event, "<Primary>Return", "<Primary>KP_Enter"):
            qltk.enqueue(self.__get_selected_songs())
            return True
        if qltk.is_accel(event, "<alt>Return"):
            songs = self.__get_selected_songs()
            if songs:
                window = SongProperties(librarian, songs, self)
                window.show()
            return True
        return False

    def __destroy(self, browser):
        self.__cover_cancel.cancel()

        self.view.bind_model(None, lambda _: None)
        self.__model_filter.destroy()
        self.__model_filter = None

        if not CoverGrid.instances():
            CoverGrid._destroy_model()

    def __cover_changed(self, manager, songs):
        songs = set(songs)

        for child in self.view:
            if not songs:
                break
            album = child.model.album
            if album is None:
                continue
            match = songs & album.songs
            if match:
                child.populate()
                songs -= match

    def __update_filter(self, scroll_up=True):
        if scroll_up:
            self.scrollwin.scroll_up()

        q = self.__search.get_query(star=["~people", "album"])
        self.__model_filter.props.filter = None if q.matches_all else q.search

    def __popup(self, widget):
        if not widget.is_selected():
            self.view.unselect_all()
        self.view.select_child(widget)

        albums = self.__get_selected_albums()
        songs = self.__get_songs_from_albums(albums)

        button_label = ngettext(
            "Reload album _cover", "Reload album _covers", len(albums)
        )
        button = MenuItem(button_label, Icons.VIEW_REFRESH)
        button.connect("activate", self.__refresh_cover, widget)

        menu = SongsMenu(self.__library, songs, items=[[button]])
        menu.show_all()
        popup_menu_at_widget(
            menu, widget, Gdk.BUTTON_SECONDARY, Gtk.get_current_event_time()
        )

    def __refresh_cover(self, menuitem, view):
        for child in self.view.get_selected_children():
            child.populate()

    def refresh_all(self):
        display_pattern = self.display_pattern
        for child in self.view:
            child.display_pattern = display_pattern

    def __get_selected_albums(self):
        items = []
        for child in self.view.get_selected_children():
            album = child.model.album
            if album is None:
                model = self.__model_filter
                return [item.album for item in model if item.album is not None]
            items.append(album)
        return items

    def __get_songs_from_albums(self, albums, sort=True):
        # Sort first by how the albums appear in the model itself,
        # then within the album using the default order.
        songs = []
        if sort:
            for album in albums:
                songs.extend(sorted(album.songs, key=lambda s: s.sort_key))
        else:
            for album in albums:
                songs.extend(album.songs)
        return songs

    def __get_selected_songs(self, sort=True):
        albums = self.__get_selected_albums()
        return self.__get_songs_from_albums(albums, sort)

    def __drag_data_get(self, view, ctx, sel, tid, etime):
        songs = self.__get_selected_songs()
        if tid == 1:
            qltk.selection_set_songs(sel, songs)
        else:
            sel.set_uris([song("~uri") for song in songs])

    def __child_activated(self, view, child):
        self.songs_activated()

    def active_filter(self, song):
        for album in self.__get_selected_albums():
            if song in album.songs:
                return True
        return False

    def can_filter_text(self):
        return True

    def filter_text(self, text):
        self.__search.set_text(text)
        if Query(text).is_parsable:
            self.__update_filter()
            self.__update_songs()

    def get_filter_text(self):
        return self.__search.get_text()

    def can_filter_albums(self):
        return True

    def can_filter(self, key):
        # Numerics are different for collections, and although title works,
        # it's not of much use here.
        if key is not None and (key.startswith("~#") or key == "title"):
            return False
        return super().can_filter(key)

    def filter_albums(self, values):
        changed = self.__select_by_func(
            lambda album: album is not None and album.key in values
        )
        self.view.grab_focus()
        if changed:
            self.__update_songs()

    def list_albums(self):
        model = self.__model_filter
        return [item.album.key for item in model if item.album is not None]

    def unfilter(self):
        self.filter_text("")

    def __select_by_func(self, func, scroll=True, one=False):
        first = True
        view = self.view
        for i, item in enumerate(self.__model_filter):
            if not func(item.album):
                continue
            child = view.get_child_at_index(i)
            if first:
                view.unselect_all()
                view.select_child(child)
                if scroll:
                    self.scrollwin.scroll_to_child(child)
                first = False
                if one:
                    break
            else:
                view.select_child(child)
        return not first

    def save(self):
        conf = self.__get_config_string()
        config.settext("browsers", "covergrid", conf)
        text = self.__search.get_text()
        config.settext("browsers", "query_text", text)

    def restore(self):
        text = config.gettext("browsers", "query_text")
        entry = self.__search
        entry.set_text(text)

        if Query(text).is_parsable:
            self.__update_filter(scroll_up=False)

        keys = config.gettext("browsers", "covergrid", "").split("\n")

        if keys != [""]:
            self.__select_by_func(
                lambda album: album is not None and album.str_key in keys
            )
        else:
            self.__select_by_func(lambda album: album is None, one=True)

    def finalize(self, restored):
        if not restored:
            self.__select_by_func(lambda album: album is None, one=True)

    def scroll(self, song):
        album_key = song.album_key
        self.__select_by_func(
            lambda album: album is not None and album.key == album_key, one=True
        )

    def activate(self):
        self.__update_songs()

    def __get_config_string(self):
        albums = []
        for child in self.view.get_selected_children():
            album = child.model.album
            if album is None:
                albums.clear()
                break
            albums.append(album)

        if not albums:
            return ""

        confval = "\n".join(a.str_key for a in albums)
        # ConfigParser strips a trailing \n so we move it to the front
        if confval and confval[-1] == "\n":
            confval = "\n" + confval[:-1]
        return confval



# Copyright 2004-2009 Joe Wreschnig, Michael Urman, Steven Robertson
#           2011-2018 Nick Boultbee
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

import re
from collections.abc import Iterable

from quodlibet.util import re_escape

DEFAULT_TAG_SPLITTERS = ["/", "&", ","]
DEFAULT_SUB_SPLITTERS = ["\u301c\u301c", "\uff08\uff09", "[]", "()", "~~", "--"]


def split_value(s, splitters=DEFAULT_TAG_SPLITTERS):
    """Splits a string. The first match in 'splitters' is used as the
    separator; subsequent matches are intentionally ignored.
    """

    def regex_for(sp):
        return r"{start}\s*{split}\s*{end}".format(
            start=r"(?:\b|(?<=\W))", split=re_escape(sp), end=r"(?:\b|(?=\W))"
        )

    if not splitters:
        return [s.strip()]
    values = s.split("\n")
    for spl in splitters:
        spl = re.compile(regex_for(spl), re.UNICODE)
        if any(spl.search(v) for v in values):
            return [st.strip() for v in values for st in spl.split(v)]
    return values


def find_subtitle(title, delimiters=DEFAULT_SUB_SPLITTERS):
    if isinstance(title, bytes):
        title = title.decode("utf-8", "replace")
    for pair in delimiters:
        if len(pair) == 2 and pair[0] in title[:-1] and title.endswith(pair[1]):
            r = len(pair[1])
            l = title[0:-r].rindex(pair[0])
            if l:
                subtitle = title[l + len(pair[0]) : -r]
                return title[:l].rstrip(), subtitle
    else:
        return title, None


def split_title(
    s, tag_splitters=DEFAULT_TAG_SPLITTERS, sub_splitters=DEFAULT_SUB_SPLITTERS
):
    title, subtitle = find_subtitle(s, sub_splitters)
    return (
        (title.strip(), split_value(subtitle, tag_splitters)) if subtitle else (s, [])
    )


__FEATURING = ["feat.", "featuring", "feat", "ft", "ft.", "with", "w/"]
__ORIGINALLY = ["originally by ", " cover"]
# Cache case-insensitive regex searches of the above
__FEAT_REGEX = [re.compile(re_escape(s + " "), re.I) for s in __FEATURING]
__ORIG_REGEX = [re.compile(re_escape(s), re.I) for s in __ORIGINALLY]


def split_people(
    s, tag_splitters=DEFAULT_TAG_SPLITTERS, sub_splitters=DEFAULT_SUB_SPLITTERS
):
    title, subtitle = find_subtitle(s, sub_splitters)
    if not subtitle:
        parts = s.split(" ")
        if len(parts) > 2:
            for feat in __FEATURING:
                try:
                    i = [p.lower() for p in parts].index(feat)
                    orig = " ".join(parts[:i])
                    others = " ".join(parts[i + 1 :])
                    return orig, split_value(others, tag_splitters)
                except (ValueError, IndexError):
                    pass
        return s, []
    old = subtitle
    # TODO: allow multiple substitutions across types, maybe
    for regex in __FEAT_REGEX + __ORIG_REGEX:
        subtitle = re.sub(regex, "", subtitle, count=1)
        if old != subtitle:
            # Only change once
            break
    values = split_value(subtitle, tag_splitters)
    return title.strip(), values


def split_album(s, sub_splitters=DEFAULT_SUB_SPLITTERS):
    name, disc = find_subtitle(s, sub_splitters)
    if not disc:
        parts = s.split(" ")
        if len(parts) > 2:
            lower = parts[-2].lower()
            if "disc" in lower or "disk" in lower:
                return " ".join(parts[:-2]), parts[-1]
        return s, None
    parts = disc.split()
    if len(parts) == 2 and parts[0].lower() in [
        "disc",
        "disk",
        "cd",
        "vol",
        "vol.",
    ]:
        try:
            return name, parts[1]
        except IndexError:
            return s, None
    else:
        return s, None


def split_genre(s: str, tag_splitters: Iterable[str] | None = None) -> list[str]:
    """Splits a single genre tag into multiple genre tags"""
    valid_split_chars = []
    for char in tag_splitters or DEFAULT_TAG_SPLITTERS:
        if char in s:
            valid_split_chars.append(char)
    if not valid_split_chars:
        return [s]
    splitchar = valid_split_chars[-1]
    # Reverses the order of DEFAULT_TAG_SPLITTERS
    # Because Genre0/Genre1, Genre2 should be split on ,
    return [genre.strip() for genre in s.split(splitchar)]



# Copyright 2016 Ryan Dellenbaugh
#           2020 Nick Boultbee
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

import os.path

from quodlibet import _, print_w
from quodlibet.plugins.query import QueryPlugin, QueryPluginError, markup_for_syntax
from quodlibet.query import Query
from quodlibet.query._match import Error as QueryError
from quodlibet import get_user_dir


class IncludeSavedSearchQuery(QueryPlugin):
    PLUGIN_ID = "include_saved"
    PLUGIN_NAME = _("Include Saved Search")
    PLUGIN_DESC = _(
        "💾 Include the results of a saved search as part of another query."
    )
    key = "saved"
    query_syntax = _("@(saved: search-name)")
    usage = markup_for_syntax(query_syntax)

    def search(self, data, body):
        return body.search(data)

    def parse_body(self, body, query_path_=None):
        if body is None:
            raise QueryPluginError
        body = body.strip().lower()
        # Use provided query file for testing
        if query_path_:
            query_path = query_path_
        else:
            query_path = os.path.join(get_user_dir(), "lists", "queries.saved")
        try:
            with open(query_path, encoding="utf-8") as query_file:
                for query_string in query_file:
                    name = next(query_file).strip().lower()
                    if name == body:
                        try:
                            return Query(query_string.strip())
                        except QueryError as e:
                            raise QueryPluginError from e
            # We've searched the whole file and haven't found a match
            print_w(f"None found for {body}")
            raise QueryPluginError
        except OSError as e:
            raise QueryPluginError from e
        except StopIteration:
            # The file has an odd number of lines. This shouldn't happen unless
            # it has been externally modified
            raise QueryPluginError from None



# Copyright 2004-2005 Joe Wreschnig, Michael Urman
#           2011 Christoph Reiter
#           2016 Ryan Dellenbaugh
#        2016-22 Nick Boultbee
#           2018 Peter Strulo
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

from __future__ import annotations

import operator
import time
from enum import auto, Enum
from numbers import Real
from typing import TypeVar
from collections.abc import Iterable

from quodlibet.formats import FILESYSTEM_TAGS, TIME_TAGS
from quodlibet.formats._audio import SIZE_TAGS, DURATION_TAGS
from quodlibet.unisearch import compile
from quodlibet.util import parse_date
from senf import fsn2text, fsnative

T = TypeVar("T")


class Error(ValueError):
    pass


class ParseError(Error):
    pass


class Node:
    def search(self, data: T) -> bool:
        raise NotImplementedError

    def filter(self, sequence: Iterable[T]) -> list[T]:
        return [s for s in sequence if self.search(s)]

    def _unpack(self) -> Node:
        return self

    def __or__(self, other: Node) -> Node:
        return NotImplemented

    def __and__(self, other: Node) -> Node:
        return NotImplemented

    def __neg__(self) -> Node:
        return Neg(self._unpack())

    @property
    def valid(self) -> bool:
        return True


class Regex(Node):
    def __init__(self, pattern: str, mod_string: str):
        self.pattern = str(pattern)
        self.mod_string = str(mod_string)

        ignore_case = "c" not in self.mod_string or "i" in self.mod_string
        dot_all = "s" in self.mod_string
        asym = "d" in self.mod_string
        try:
            re = compile(self.pattern, ignore_case, dot_all, asym)
            self.search = re  # type: ignore
        except ValueError as e:
            raise ParseError(
                f"The regular expression /{self.pattern}/ is invalid."
            ) from e

    def __repr__(self):
        return f"<Regex pattern={self.pattern} mod={self.mod_string}>"


class True_(Node):
    """Always True"""

    def search(self, data):
        return True

    def filter(self, sequence):
        return list(sequence)

    def __repr__(self):
        return "<True>"

    def __or__(self, other):
        return self

    def __and__(self, other):
        return other._unpack()


class False_(Node):
    """Always False"""

    def search(self, data):
        return False

    def filter(self, sequence):
        return []

    def __repr__(self):
        return "<False>"

    def __or__(self, other):
        return other._unpack()

    def __and__(self, other):
        return self


class Union(Node):
    """True if the object matches any of its REs."""

    def __init__(self, res: list[Node]):
        self.res = res

    def search(self, data):
        for re in self.res:
            if re.search(data):
                return True
        return False

    def __repr__(self):
        return f"<Union {self.res!r}>"

    def __or__(self, other):
        other = other._unpack()

        if isinstance(other, Union):
            return Union(self.res + list(other.res))
        if isinstance(other, True_):
            return other.__or__(self)

        return Union(self.res + [other])

    def __and__(self, other):
        other = other._unpack()

        if isinstance(other, Inter | True_):
            return other.__and__(self)

        return Inter([self, other])


class Inter(Node):
    """True if the object matches all of its REs."""

    def __init__(self, res):
        self.res = res

    def search(self, data):
        for re in self.res:
            if not re.search(data):
                return False
        return True

    def filter(self, sequence):
        current = sequence
        for re in self.res:
            current = filter(re.search, current)
        if not isinstance(current, list):
            current = list(current)
        return current

    def __repr__(self):
        return f"<Inter {self.res!r}>"

    def __and__(self, other):
        other = other._unpack()

        if isinstance(other, Inter):
            return Inter(self.res + other.res)

        if isinstance(other, True_):
            return other.__and__(self)

        return Inter(self.res + [other])

    def __or__(self, other):
        other = other._unpack()
        if isinstance(other, Union | True_):
            return other.__or__(self)
        return Union([self, other])


class Neg(Node):
    """True if the object doesn't match its RE."""

    def __init__(self, res):
        self.res = res

    def search(self, data):
        return not self.res.search(data)

    def __repr__(self):
        return f"<Neg {self.res!r}>"

    def __and__(self, other):
        other = other._unpack()
        if isinstance(other, True_):
            return other.__and__(self)
        return Inter([self, other])

    def __or__(self, other):
        other = other._unpack()
        if isinstance(other, True_):
            return other.__or__(self)
        return Union([self, other])

    def __neg__(self):
        return self.res


class Numcmp(Node):
    """Numeric comparisons"""

    operators = {
        "<": operator.lt,
        "<=": operator.le,
        ">": operator.gt,
        ">=": operator.ge,
        "=": operator.eq,
        "==": operator.eq,
        "!=": operator.ne,
    }

    def __init__(self, expr: Numexpr, op: str, expr2: Numexpr):
        self._expr = expr
        self._op = self.operators[op]
        self._expr2 = expr2
        units = expr2.units()

        if units and not expr.valid_for_units(units):
            raise ParseError(f"Wrong units for {expr}")

    def search(self, data):
        time_ = time.time()
        use_date = self._expr.use_date() or self._expr2.use_date()
        val = self._expr.evaluate(data, time_, use_date)
        val2 = self._expr2.evaluate(data, time_, use_date)
        if val is not None and val2 is not None:
            return self._op(val, val2)
        return False

    def __repr__(self):
        return (
            f"<Numcmp expr={self._expr!r}, "
            f"op={self._op.__name__!r}, "
            f"expr2={self._expr2!r}>"
        )

    def __and__(self, other):
        other = other._unpack()
        if isinstance(other, True_):
            return other.__and__(self)
        return Inter([self, other])

    def __or__(self, other):
        other = other._unpack()
        if isinstance(other, True_):
            return other.__or__(self)
        return Union([self, other])


class Numexpr:
    """Expression in numeric comparison"""

    def evaluate(self, data: T, time: float, use_date: bool):
        """Evaluate the expression to a number.
        :param data: is the audiofile to evaluate for
        :param time: is the current time
        :param use_date: whether to evaluate as a date
                         (used to handle expressions like 2015-02-11
                          that look like dates and subtraction)
        """
        raise NotImplementedError

    def use_date(self) -> bool:
        """Returns whether to force the final comparison to compare the date
        values instead of the number values."""
        return False

    def units(self) -> Units | None:
        """Returns optional (converted) units for this number"""
        return None

    def valid_for_units(self, units: Units) -> bool:
        """Returns true if the given unit is valid for this expression"""
        return True


class NumexprTag(Numexpr):
    """Numeric tag"""

    def __init__(self, tag: str):
        self._tag = tag
        self._ftag = "~#" + self._tag
        # Strip aggregate function from tag
        self._base_ftag = (self._ftag or "").split(":")[0]

    def valid_for_units(self, units: Units) -> bool:
        return self._base_ftag in UNITS_TO_TAGS[units]

    def evaluate(self, data, time, use_date):
        if self._tag == "date":
            date = data("date")
            if not date:
                return None
            try:
                num = parse_date(date)
            except ValueError:
                return None
        else:
            num = data(self._ftag, None)
        if num is not None:
            if self._base_ftag in TIME_TAGS:
                num = time - num
            return round(num, 2)
        return None

    def __repr__(self):
        return f"<NumexprTag tag={self._tag!r}>"

    def use_date(self):
        return self._tag == "date"


class NumexprUnary(Numexpr):
    """Unary numeric operation (like -)"""

    operators = {"-": operator.neg}

    def __init__(self, op: str, expr: Numexpr):
        self.__op = self.operators[op]
        self.__expr = expr

    def evaluate(self, data, time, use_date):
        val = self.__expr.evaluate(data, time, use_date)
        if val is not None:
            return self.__op(val)
        return None

    def __repr__(self):
        return f"<NumexprUnary op={self.__op!r} expr={self.__expr!r}>"

    def use_date(self):
        return self.__expr.use_date()


class NumexprBinary(Numexpr):
    """Binary numeric operation (like + or *)"""

    operators = {
        "-": operator.sub,
        "+": operator.add,
        "*": operator.mul,
        "/": operator.floordiv,
    }

    precedence = {
        operator.sub: 1,
        operator.add: 1,
        operator.mul: 2,
        operator.floordiv: 2,
    }

    def __init__(self, op: str, expr: Numexpr, expr2: Numexpr):
        self.__op = self.operators[op]
        self.__expr = expr
        self.__expr2 = expr2
        # Rearrange expressions for operator precedence
        if (
            isinstance(expr, NumexprBinary)
            and self.precedence[expr.__op] < self.precedence[self.__op]
        ):
            self.__expr = expr.__expr
            self.__op = expr.__op
            expr.__expr = expr.__expr2
            expr.__op = self.operators[op]
            expr.__expr2 = expr2
            self.__expr2 = expr

    def evaluate(self, data, time, use_date):
        val = self.__expr.evaluate(data, time, use_date)
        val2 = self.__expr2.evaluate(data, time, use_date)
        if val is not None and val2 is not None:
            try:
                return self.__op(val, val2)
            except ZeroDivisionError:
                return val * float("inf")
        return None

    def __repr__(self):
        return (
            f"<NumexprBinary "
            f"op={self.__op!r} "
            f"expr={self.__expr!r} "
            f"expr2={self.__expr2!r}>"
        )

    def use_date(self):
        return self.__expr.use_date() or self.__expr2.use_date()


class NumexprGroup(Numexpr):
    """Parenthesized group in numeric expression"""

    def __init__(self, expr: Numexpr):
        self.__expr = expr

    def evaluate(self, data, time, use_date):
        return self.__expr.evaluate(data, time, use_date)

    def __repr__(self):
        return f"<NumexprGroup expr={self.__expr!r}>"

    def use_date(self):
        return self.__expr.use_date()


class NumexprNumber(Numexpr):
    """Number in numeric expression"""

    def __init__(self, value: Real, units: Units | None = None):
        self._value = float(value)
        self._units = units

    def evaluate(self, data, time, use_date):
        return self._value

    def units(self) -> Units | None:
        return self._units

    def __repr__(self):
        return f"<NumexprNumber value={self._value:.2f}>"


class NumexprNow(Numexpr):
    """Current time, with optional offset"""

    def __init__(self, offset=0):
        self.__offset = offset

    def evaluate(self, data, time, use_date):
        return time - self.__offset

    def __repr__(self):
        return f"<NumexprNow offset={self.__offset!r}>"


class NumexprNumberOrDate(Numexpr):
    """An ambiguous value like 2015-09-25 than can be interpreted as either
    a number or a date."""

    def __init__(self, date):
        self.date = parse_date(date)
        parts = date.split("-")
        self.number = int(parts[0])
        if len(parts) > 1:
            self.number -= int(parts[1])
        if len(parts) > 2:
            self.number -= int(parts[2])

    def evaluate(self, data, time, use_date):
        if use_date:
            return self.date
        return self.number

    def __repr__(self):
        return f"<NumexprNumberOrDate number={self.number!r} date={self.date!r}>"


class Units(Enum):
    SECONDS = auto()
    BYTES = auto()


UNITS_TO_TAGS = {Units.SECONDS: TIME_TAGS | DURATION_TAGS, Units.BYTES: SIZE_TAGS}


def numexprUnit(value, unit):
    """Process numeric units and return NumexprNumber"""

    unit = unit.lower().strip()
    converted = Units.SECONDS
    # Time units
    if unit.startswith("second"):
        value = value
    elif unit.startswith("minute"):
        value *= 60
    elif unit.startswith("hour"):
        value *= 60 * 60
    elif unit.startswith("day"):
        value *= 24 * 60 * 60
    elif unit.startswith("week"):
        value *= 7 * 24 * 60 * 60
    elif unit.startswith("month"):
        value *= 30 * 24 * 60 * 60
    elif unit.startswith("year"):
        value *= 365 * 24 * 60 * 60
    # Size units
    elif unit.startswith("g"):
        value *= 1024**3
        converted = Units.BYTES
    elif unit.startswith("m"):
        value *= 1024**2
        converted = Units.BYTES
    elif unit.startswith("k"):
        value *= 1024
        converted = Units.BYTES
    elif unit.startswith("b"):
        converted = Units.SECONDS
    elif unit:
        raise ParseError(f"No such unit: {unit!r}")
    return NumexprNumber(value, converted)


def numexprTagOrSpecial(tag):
    """Handle special values that look like tags"""

    if tag == "now":
        return NumexprNow()
    if tag == "today":
        return NumexprNow(offset=24 * 60 * 60)
    return NumexprTag(tag)


class Tag(Node):
    """See if a property of the object matches its RE."""

    # Shorthand for common tags.
    ABBRS = {
        "a": "artist",
        "b": "album",
        "v": "version",
        "t": "title",
        "n": "tracknumber",
        "d": "date",
    }

    def __init__(self, names: Iterable[str], res):
        self.res = res
        self._names = []
        self.__intern = []
        self.__fs = []

        names = [Tag.ABBRS.get(n.lower(), n.lower()) for n in names]
        for name in names:
            if name[:1] == "~":
                if name.startswith("~#"):
                    raise ValueError("numeric tags not supported")
                if name in FILESYSTEM_TAGS:
                    self.__fs.append(name)
                else:
                    self.__intern.append(name)
            else:
                self._names.append(name)

    def search(self, data):
        search = self.res.search
        fs_default = fsnative()

        for name in self._names:
            val = data.get(name)
            if val is None:
                if name in ("filename", "mountpoint"):
                    val = fsn2text(data.get("~" + name, fs_default))
                else:
                    val = data.get("~" + name, "")

            if search(val):
                return True

        for name in self.__intern:
            if search(data(name)):
                return True

        for name in self.__fs:
            if search(fsn2text(data(name, fs_default))):
                return True

        return False

    def __repr__(self):
        names = self._names + self.__intern
        return f"<Tag names={names!r}, res={self.res!r}>"

    def __and__(self, other):
        other = other._unpack()
        if isinstance(other, True_):
            return other.__and__(self)
        return Inter([self, other])

    def __or__(self, other):
        other = other._unpack()
        if isinstance(other, True_):
            return other.__or__(self)
        return Union([self, other])


class Extension(Node):
    """Plugin-defined query extension

    Syntax is @(plugin_name) or @(plugin_name: body)

    Raises a ParseError if no plugin is loaded for the name, or if the plugin
    fails to parse the body"""

    def __init__(self, name, body):
        # pulls in gtk+
        from quodlibet.plugins.query import QUERY_HANDLER, QueryPluginError

        self.__name = name
        self.__valid = True
        self.__body = body

        try:
            self.__plugin = QUERY_HANDLER.get_plugin(name)
        except KeyError:
            self.__valid = False
            return

        try:
            self.__body = self.__plugin.parse_body(body)
        except QueryPluginError:
            self.__valid = False
            return

    def search(self, data):
        return self.__valid and self.__plugin.search(data, self.__body)

    @property
    def valid(self) -> bool:
        return self.__valid

    def __repr__(self):
        return (
            f"<Extension "
            f"name={self.__name!r} valid={self.__valid!r} body={self.__body!r}>"
        )

    def __and__(self, other):
        other = other._unpack()

        if isinstance(other, Inter | True_):
            return other.__and__(self)

        return Inter([self] + [other])

    def __or__(self, other):
        other = other._unpack()

        if isinstance(other, Union | True_):
            return other.__or__(self)

        return Union([self, other])



# Copyright 2015 Christoph Reiter
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

"""Varisour function for figuring out which platform wa are running on
and under which environment.
"""

import os
import sys
import ctypes
import configparser
import fnmatch

from gi.repository import GLib, Gio


def _dbus_name_owned(name):
    """Returns True if the dbus name has an owner"""

    if not is_linux():
        return False

    return dbus_name_owned(name)


def is_flatpak():
    """If we are running in a flatpak"""

    return is_linux() and os.path.exists("/.flatpak-info")


def matches_flatpak_runtime(pattern: str) -> bool:
    """Pass a fnmatch pattern for matching the flatpak runtime ID"""

    if not is_linux():
        return False

    config = configparser.ConfigParser()
    try:
        with open("/.flatpak-info", encoding="utf-8") as f:
            config.read_file(f)
        runtime = config.get("Application", "runtime")
    except (OSError, configparser.Error):
        return False

    return fnmatch.fnmatchcase(runtime, pattern)


def is_plasma():
    """If we are running under plasma"""

    return _dbus_name_owned("org.kde.plasmashell")


def is_unity():
    """If we are running under Ubuntu/Unity"""

    return _dbus_name_owned("com.canonical.Unity.Launcher")


def is_enlightenment():
    """If we are running under Enlightenment"""

    return _dbus_name_owned("org.enlightenment.wm.service")


def is_linux():
    """If we are on Linux (or similar)"""

    return not is_windows() and not is_osx()


def is_windows():
    """If we are running under Windows or Wine"""

    return os.name == "nt"


def is_wine():
    """If we are running under Wine"""

    if not is_windows():
        return False

    try:
        ctypes.cdll.ntdll.wine_get_version  # noqa
    except AttributeError:
        return False
    else:
        return True


def is_osx():
    """If we are running under OS X"""

    return sys.platform == "darwin"


def dbus_name_owned(name):
    """Returns True if the dbus name has an owner"""

    BUS_DAEMON_NAME = "org.freedesktop.DBus"  # noqa
    BUS_DAEMON_PATH = "/org/freedesktop/DBus"  # noqa
    BUS_DAEMON_IFACE = "org.freedesktop.DBus"  # noqa

    try:
        bus = Gio.DBusProxy.new_for_bus_sync(
            Gio.BusType.SESSION,
            Gio.DBusProxyFlags.NONE,
            None,
            BUS_DAEMON_NAME,
            BUS_DAEMON_PATH,
            BUS_DAEMON_IFACE,
            None,
        )
        return bus.NameHasOwner("(s)", name)
    except GLib.Error:
        return False



# Copyright 2018 Jan Korte
#           2020 Daniel Petrescu
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

import os
import shutil
import unicodedata
from pathlib import Path

from gi.repository import Gtk, Pango
from senf import fsn2text

from quodlibet import _
from quodlibet import app
from quodlibet import config
from quodlibet import get_user_dir
from quodlibet import ngettext
from quodlibet import qltk
from quodlibet import util
from quodlibet.pattern import FileFromPattern
from quodlibet.plugins import PM
from quodlibet.plugins import PluginConfigMixin
from quodlibet.plugins.events import EventPlugin
from quodlibet.qltk import Icons
from quodlibet.qltk.cbes import ComboBoxEntrySave
from quodlibet.qltk.ccb import ConfigCheckButton
from quodlibet.qltk.views import HintedTreeView
from quodlibet.query import Query
from quodlibet.util import print_d, print_e, print_exc
from quodlibet.util.enum import enum
from quodlibet.util.path import strip_win32_incompat_from_path

PLUGIN_CONFIG_SECTION = "synchronize_to_device"


class Entry:
    """
    An entry in the tree of previewed export paths.
    """

    @enum
    class Tags(str):
        """
        Various tags that will be used in the output.
        """

        EMPTY = ""
        PENDING_COPY = _("Pending copy")
        PENDING_DELETE = _("Pending delete")
        DELETE = _("delete")
        SKIP = _("Skip")
        SKIP_DUPLICATE = _("DUPLICATE")
        IN_PROGRESS_SYNC = _("Synchronizing")
        IN_PROGRESS_DELETE = _("Deleting")
        RESULT_SUCCESS = _("Success")
        RESULT_FAILURE = _("FAILURE")
        RESULT_SKIP_EXISTING = _("Skipped existing file")

    def __init__(self, song, export_path=None):
        self._song = song
        self.export_path = export_path or ""
        self.tag = self.Tags.EMPTY
        self._filename = None

    @property
    def filename(self):
        if self._song is not None:
            return fsn2text(self._song("~filename"))
        return self._filename

    @filename.setter
    def filename(self, name):
        if self._song is None:
            self._filename = name
        else:
            raise ValueError(_("Cannot set the filename of a song."))


class SyncToDevice(EventPlugin, PluginConfigMixin):
    PLUGIN_ICON = Icons.NETWORK_TRANSMIT
    PLUGIN_ID = PLUGIN_CONFIG_SECTION
    PLUGIN_NAME = _("Synchronize to Device")
    PLUGIN_DESC = _(
        "Synchronizes all songs from the selected saved searches "
        "with the specified folder."
    )

    CONFIG_SECTION = PLUGIN_CONFIG_SECTION
    CONFIG_QUERY_PREFIX = "query_"
    CONFIG_PATH_KEY = "{}_{}".format(PLUGIN_CONFIG_SECTION, "path")
    CONFIG_PATTERN_KEY = "{}_{}".format(PLUGIN_CONFIG_SECTION, "pattern")

    path_query = os.path.join(get_user_dir(), "lists", "queries.saved")
    path_pattern = os.path.join(get_user_dir(), "lists", "renamepatterns")

    spacing_main = 20
    spacing_large = 6
    spacing_small = 3
    summary_sep = " " * 2
    summary_sep_list = "," + summary_sep

    default_export_pattern = os.path.join("<artist>", "<album>", "<title>")

    model_cols = {
        "entry": (0, object),
        "tag": (1, str),
        "filename": (2, str),
        "export": (3, str),
    }

    def PluginPreferences(self, parent):
        # Check if the queries file exists
        if not os.path.exists(self.path_query):
            return self._no_queries_frame()

        # Read saved searches from file
        self.queries = {}
        with open(self.path_query, encoding="utf-8") as query_file:
            for query_string in query_file:
                name = next(query_file).strip()
                self.queries[name] = Query(query_string.strip())
        if not self.queries:
            # query_file is empty
            return self._no_queries_frame()

        main_vbox = Gtk.VBox(spacing=self.spacing_main)
        self.main_vbox = main_vbox

        # Saved search selection frame
        saved_search_vbox = Gtk.VBox(spacing=self.spacing_large)
        self.saved_search_vbox = saved_search_vbox
        for query_name, _query in self.queries.items():
            query_config = self.CONFIG_QUERY_PREFIX + query_name
            check_button = ConfigCheckButton(
                query_name, PM.CONFIG_SECTION, self._config_key(query_config)
            )
            check_button.set_active(self.config_get_bool(query_config))
            saved_search_vbox.pack_start(check_button, False, False, 0)
        saved_search_scroll = self._expandable_scroll(min_h=0, max_h=300)
        saved_search_scroll.add(saved_search_vbox)
        frame = qltk.Frame(
            label=_("Synchronize the following saved searches:"),
            child=saved_search_scroll,
        )
        main_vbox.pack_start(frame, False, False, 0)

        # Destination path entry field
        destination_entry = Gtk.Entry(
            placeholder_text=_("The absolute path to your export location"),
            text=config.get(PM.CONFIG_SECTION, self.CONFIG_PATH_KEY, ""),
        )
        destination_entry.connect("changed", self._destination_path_changed)
        self.destination_entry = destination_entry

        # Destination path selection button
        destination_button = qltk.Button(label="", icon_name=Icons.FOLDER_OPEN)
        destination_button.connect("clicked", self._select_destination_path)

        # Destination path hbox
        destination_path_hbox = Gtk.HBox(spacing=self.spacing_small)
        destination_path_hbox.pack_start(destination_entry, True, True, 0)
        destination_path_hbox.pack_start(destination_button, False, False, 0)

        # Destination path information
        destination_warn_label = self._label_with_icon(
            _(
                "All pre-existing files in the destination folder that aren't in "
                "the saved searches will be deleted."
            ),
            Icons.DIALOG_WARNING,
        )
        destination_info_label = self._label_with_icon(
            _(
                "For devices mounted with MTP, export to a local destination "
                "folder, then transfer it to your device with rsync. "
                "Or, when syncing many files to an Android Device, use adb-sync, "
                "which is much faster."
            ),
            Icons.DIALOG_INFORMATION,
        )

        # Destination path frame
        destination_vbox = Gtk.VBox(spacing=self.spacing_large)
        destination_vbox.pack_start(destination_path_hbox, False, False, 0)
        destination_vbox.pack_start(destination_warn_label, False, False, 0)
        destination_vbox.pack_start(destination_info_label, False, False, 0)
        frame = qltk.Frame(label=_("Destination path:"), child=destination_vbox)
        main_vbox.pack_start(frame, False, False, 0)

        # Export pattern frame
        export_pattern_combo = ComboBoxEntrySave(
            self.path_pattern,
            [self.default_export_pattern],
            title=_("Path Patterns"),
            edit_title=_("Edit saved patterns…"),
        )
        export_pattern_combo.enable_clear_button()
        export_pattern_combo.show_all()
        export_pattern_entry = export_pattern_combo.get_child()
        export_pattern_entry.set_placeholder_text(
            _("The structure of the exported filenames, based on their tags")
        )
        export_pattern_entry.set_text(
            config.get(
                PM.CONFIG_SECTION, self.CONFIG_PATTERN_KEY, self.default_export_pattern
            )
        )
        export_pattern_entry.connect("changed", self._export_pattern_changed)
        self.export_pattern_entry = export_pattern_entry
        frame = qltk.Frame(label=_("Export pattern:"), child=export_pattern_combo)
        main_vbox.pack_start(frame, False, False, 0)

        # Start preview button
        preview_start_button = qltk.Button(
            label=_("Preview"), icon_name=Icons.VIEW_REFRESH
        )
        preview_start_button.set_visible(True)
        preview_start_button.connect("clicked", self._start_preview)
        self.preview_start_button = preview_start_button

        # Stop preview button
        preview_stop_button = qltk.Button(
            label=_("Stop preview"), icon_name=Icons.PROCESS_STOP
        )
        preview_stop_button.set_visible(False)
        preview_stop_button.set_no_show_all(True)
        preview_stop_button.connect("clicked", self._stop_preview)
        self.preview_stop_button = preview_stop_button

        # Details view
        column_types = [column[1] for column in self.model_cols.values()]
        self.model = Gtk.ListStore(*column_types)
        self.details_tree = details_tree = HintedTreeView(model=self.model)
        details_scroll = self._expandable_scroll()
        details_scroll.set_shadow_type(Gtk.ShadowType.IN)
        details_scroll.add(details_tree)
        self.renders = {}

        # Preview column: status
        render = Gtk.CellRendererText()
        column = self._tree_view_column(
            render,
            self._cdf_status,
            title=_("Status"),
            expand=False,
            sort=self._model_col_id("tag"),
        )
        details_tree.append_column(column)

        # Preview column: file
        render = Gtk.CellRendererText()
        column = self._tree_view_column(
            render,
            self._cdf_source_path,
            title=_("Source File"),
            sort=self._model_col_id("filename"),
        )
        details_tree.append_column(column)

        # Preview column: export path
        render = Gtk.CellRendererText()
        render.set_property("editable", True)
        render.connect("edited", self._row_edited)
        column = self._tree_view_column(
            render,
            self._cdf_export_path,
            title=_("Export Path"),
            sort=self._model_col_id("export"),
        )
        details_tree.append_column(column)

        # Status labels
        self.status_operation = Gtk.Label(
            xalign=0.0, yalign=0.5, wrap=True, visible=False, no_show_all=True
        )
        self.status_progress = Gtk.Label(
            xalign=0.0, yalign=0.5, wrap=True, visible=False, no_show_all=True
        )
        self.status_duplicates = self._label_with_icon(
            _(
                "Duplicate export paths detected! The export paths above can be "
                "edited before starting the synchronization."
            ),
            Icons.DIALOG_WARNING,
            visible=False,
        )
        self.status_deletions = self._label_with_icon(
            _(
                "Existing files in the destination path will be deleted (except "
                "files named 'cover.jpg')!"
            ),
            Icons.DIALOG_WARNING,
            visible=False,
        )

        # Section for previewing exported files
        preview_vbox = Gtk.VBox(spacing=self.spacing_large)
        preview_vbox.pack_start(preview_start_button, False, False, 0)
        preview_vbox.pack_start(preview_stop_button, False, False, 0)
        preview_vbox.pack_start(details_scroll, True, True, 0)
        preview_vbox.pack_start(self.status_operation, False, False, 0)
        preview_vbox.pack_start(self.status_progress, False, False, 0)
        preview_vbox.pack_start(self.status_duplicates, False, False, 0)
        preview_vbox.pack_start(self.status_deletions, False, False, 0)
        main_vbox.pack_start(preview_vbox, True, True, 0)

        # Start sync button
        sync_start_button = qltk.Button(
            label=_("Start synchronization"), icon_name=Icons.DOCUMENT_SAVE
        )
        sync_start_button.set_sensitive(False)
        sync_start_button.set_visible(True)
        sync_start_button.connect("clicked", self._start_sync)
        self.sync_start_button = sync_start_button

        # Stop sync button
        sync_stop_button = qltk.Button(
            label=_("Stop synchronization"), icon_name=Icons.PROCESS_STOP
        )
        sync_stop_button.set_visible(False)
        sync_stop_button.set_no_show_all(True)
        sync_stop_button.connect("clicked", self._stop_sync)
        self.sync_stop_button = sync_stop_button

        # Section for the sync buttons
        sync_vbox = Gtk.VBox(spacing=self.spacing_large)
        sync_vbox.pack_start(sync_start_button, False, False, 0)
        sync_vbox.pack_start(sync_stop_button, False, False, 0)
        main_vbox.pack_start(sync_vbox, False, False, 0)

        return main_vbox

    @staticmethod
    def _no_queries_frame():
        """
        Create a frame to use when there are no saved searches.

        :return: A new Frame.
        """
        return qltk.Frame(_("No saved searches yet, create some and come back!"))

    def _expandable_scroll(self, min_h=50, max_h=-1, expand=True):
        """
        Create a ScrolledWindow that expands as content is added.

        :param min_h: The minimum height of the window, in pixels.
        :param max_h: The maximum height of the window, in pixels. It will grow
                      up to this height before it starts scrolling the content.
        :param expand: Whether the window should expand.
        :return: A new ScrolledWindow.
        """
        return Gtk.ScrolledWindow(
            min_content_height=min_h,
            max_content_height=max_h,
            propagate_natural_height=expand,
        )

    def _label_with_icon(self, text, icon_name, visible=True):
        """
        Create a new label with an icon to the left of the text.

        :param text:      The new text to set for the label.
        :param icon_name: An icon name or None.
        :return: A HBox containing an icon followed by a label.
        """
        image = Gtk.Image.new_from_icon_name(icon_name, Gtk.IconSize.BUTTON)
        label = Gtk.Label(label=text, xalign=0.0, yalign=0.5, wrap=True)

        hbox = Gtk.HBox(spacing=self.spacing_large)
        if not visible:
            hbox.set_visible(False)
            hbox.set_no_show_all(True)
        hbox.pack_start(image, False, False, 0)
        hbox.pack_start(label, True, True, 0)

        return hbox

    def _tree_view_column(
        self, render, cdf, title=None, sort=None, expand=True, resize=True, reorder=True
    ):
        """
        Create a new TreeViewColumn with the given properties.

        :param render:  The A Gtk.CellRenderer of this cell.
        :param cdf:     The Gtk.TreeCellDataFunc to use for updating content.
        :param title:   The column's title.
        :param sort:    The model column to use when sorting this column.
        :param expand:  Whether the column width should automatically expand.
        :param resize:  Whether the column can be resized.
        :param reorder: Whether the column can be reordered.
        :return: The new TreeViewColumn.
        """
        tvc = Gtk.TreeViewColumn()
        tvc.set_sizing(Gtk.TreeViewColumnSizing.AUTOSIZE)
        tvc.set_expand(expand)
        tvc.set_resizable(resize)
        tvc.set_reorderable(reorder)
        if title:
            tvc.set_title(title)
        if resize:
            render.set_property("ellipsize", Pango.EllipsizeMode.END)
        if sort:
            tvc.set_sort_column_id(sort)
        tvc.set_cell_data_func(render, cdf)
        tvc.pack_start(render, True)
        self.renders[tvc] = render
        return tvc

    def _destination_path_changed(self, entry):
        """
        Save the destination path to the global config when the path changes.

        :param entry: The destination path entry field.
        """
        config.set(PM.CONFIG_SECTION, self.CONFIG_PATH_KEY, entry.get_text())

    def _select_destination_path(self, button):
        """
        Show a folder selection dialog to select the destination path
        from the file system.

        :param button: The destination path selection button.
        """
        dialog = Gtk.FileChooserDialog(
            title=_("Choose destination path"),
            action=Gtk.FileChooserAction.SELECT_FOLDER,
            select_multiple=False,
            create_folders=True,
            local_only=False,
            show_hidden=True,
        )
        dialog.add_buttons(
            _("_Cancel"), Gtk.ResponseType.CANCEL, _("_Save"), Gtk.ResponseType.OK
        )
        dialog.set_default_response(Gtk.ResponseType.OK)

        # If there is an existing path in the entry field,
        # make that path the default
        destination_entry_text = self.destination_entry.get_text()
        if destination_entry_text != "":
            dialog.set_current_folder(destination_entry_text)

        # Show the dialog and get the selected path
        response = dialog.run()
        response_path = dialog.get_filename()

        # Close the dialog and save the selected path
        dialog.destroy()
        if response == Gtk.ResponseType.OK and response_path != destination_entry_text:
            self.destination_entry.set_text(response_path)

    def _export_pattern_changed(self, entry):
        """
        Save the export pattern to the global config when the pattern changes.

        :param entry: The export pattern entry field.
        """
        config.set(PM.CONFIG_SECTION, self.CONFIG_PATTERN_KEY, entry.get_text())

    def _cdf_status(self, column, cell, model, iter_, data):
        """
        Handle entering data into the "Status" column of the sync previews.
        """
        cell.set_property("text", model[iter_][self._model_col_id("tag")])

    def _cdf_source_path(self, column, cell, model, iter_, data):
        """
        Handle entering data into the "File" column of the sync previews.
        """
        cell.set_property("text", model[iter_][self._model_col_id("filename")])

    def _cdf_export_path(self, column, cell, model, iter_, data):
        """
        Handle entering data into the "Export" column of the sync previews.
        """
        cell.set_property("text", model[iter_][self._model_col_id("export")])

    def _row_edited(self, renderer, path, entered_path):
        """
        Handle a manual edit of a previewed export path.

        :param renderer:        The object which received the signal.
        :param path:            The path identifying the edited cell.
        :param entered_path:    The new path entered by the user.
        """

        def _update_warnings():
            """
            Toggle the visibility of the status warning labels based on the song
            counts.
            """
            if self.c_song_dupes == 0:
                self.status_duplicates.set_visible(False)
            else:
                self.status_duplicates.set_visible(True)

            if self.c_songs_delete == 0:
                self.status_deletions.set_visible(False)
            else:
                self.status_deletions.set_visible(True)

        def _make_duplicate(entry, old_unique):
            """Mark the given entry as a duplicate."""
            print_d(entry.filename)
            entry.tag = Entry.Tags.SKIP_DUPLICATE
            self.c_song_dupes += 1
            if old_unique:
                self.c_songs_copy -= 1
            _update_warnings()

        def _make_unique(entry, old_duplicate):
            """Mark the given entry as a unique file."""
            print_d(entry.filename)
            entry.tag = Entry.Tags.PENDING_COPY
            self.c_songs_copy += 1
            if old_duplicate:
                self.c_song_dupes -= 1
            _update_warnings()

        def _make_skip(entry, counter):
            """Skip the given entry during synchronization."""
            print_d(entry.filename)
            entry.tag = Entry.Tags.SKIP
            entry.export_path = ""
            return counter - 1

        def _update_other_song(model, path, iter_, *data):
            """
            Update a previewed path based on the current change.
            This is a callback function passed to Gtk.TreeModel.foreach() to
            iterate over the rows in a tree model.

            :return: True to stop iterating, False to continue.
            """
            model_entry = model[path][self._model_col_id("entry")]
            if (
                model_entry is entry
                or model_entry.tag == Entry.Tags.DELETE
                or model_entry.export_path == ""
            ):
                pass
            elif (
                model_entry.export_path == entered_path
                and model_entry.tag == Entry.Tags.PENDING_COPY
            ):
                _make_duplicate(model_entry, True)
                self._update_model_value(iter_, "tag", model_entry.tag)
            elif (
                model_entry.tag == Entry.Tags.SKIP_DUPLICATE
                and model_entry.export_path != entered_path
                and self._get_paths()[model_entry.export_path] == 1
            ):
                _make_unique(model_entry, True)
                self._update_model_value(iter_, "tag", model_entry.tag)
            return False

        path = Gtk.TreePath.new_from_string(path)
        entry = self.model[path][self._model_col_id("entry")]
        if entry.export_path != entered_path:
            old_path, new_path = {}, {}

            old_path["duplicate"] = entry.tag == Entry.Tags.SKIP_DUPLICATE
            old_path["delete"] = entry.tag == Entry.Tags.PENDING_DELETE
            old_path["empty"] = not entry.export_path and not old_path["delete"]
            old_path["unique"] = not (
                old_path["duplicate"] or old_path["delete"] or old_path["empty"]
            )
            old_path_inv = {}
            for key, value in old_path.items():
                old_path_inv.setdefault(value, []).append(key)

            previewed_paths = self._get_paths().keys()

            new_path["duplicate"] = entered_path in previewed_paths
            new_path["delete"] = entered_path.lower() == Entry.Tags.DELETE
            new_path["empty"] = not entered_path and not new_path["delete"]
            new_path["unique"] = not (
                new_path["duplicate"] or new_path["delete"] or new_path["empty"]
            )
            new_path_inv = {}
            for key, value in new_path.items():
                new_path_inv.setdefault(value, []).append(key)

            print_d(
                _(
                    "Export path changed from [{old_path}] to [{new_path}] "
                    "for file [{filename}]"
                ).format(
                    filename=entry.filename,
                    old_path=" ".join(old_path_inv[True]),
                    new_path=" ".join(new_path_inv[True]),
                )
            )

            # If the old path was empty...
            if old_path["empty"] and new_path["empty"]:
                pass
            elif old_path["empty"] and new_path["delete"]:
                try:
                    Path(entry.filename).relative_to(self.expanded_destination)
                    entry.tag = Entry.Tags.PENDING_DELETE
                    self.c_songs_delete += 1
                    _update_warnings()
                except ValueError:
                    pass
            elif old_path["empty"] and new_path["duplicate"]:
                _make_duplicate(entry, False)
                entry.export_path = entered_path
            elif old_path["empty"] and new_path["unique"]:
                _make_unique(entry, False)
                entry.export_path = entered_path

            # If the old path was a deletion...
            elif old_path["delete"] and new_path["empty"]:
                pass
            elif old_path["delete"] and new_path["delete"]:
                self.c_songs_delete = _make_skip(entry, self.c_songs_delete)
                _update_warnings()
            elif old_path["delete"] and new_path["duplicate"]:
                pass
            elif old_path["delete"] and new_path["unique"]:
                pass

            # If the old path was a duplicate...
            elif old_path["duplicate"] and new_path["empty"]:
                self.c_song_dupes = _make_skip(entry, self.c_song_dupes)
                self.model.foreach(_update_other_song)
                _update_warnings()
            elif old_path["duplicate"] and new_path["delete"]:
                self.c_song_dupes = _make_skip(entry, self.c_song_dupes)
                self.model.foreach(_update_other_song)
                _update_warnings()
            elif old_path["duplicate"] and new_path["duplicate"]:
                entry.export_path = entered_path
            elif old_path["duplicate"] and new_path["unique"]:
                _make_unique(entry, True)
                entry.export_path = entered_path
                self.model.foreach(_update_other_song)

            # If the old path was unique...
            elif old_path["unique"] and new_path["empty"]:
                self.c_songs_copy = _make_skip(entry, self.c_songs_copy)
                self.model.foreach(_update_other_song)
                _update_warnings()
            elif old_path["unique"] and new_path["delete"]:
                self.c_songs_copy = _make_skip(entry, self.c_songs_copy)
                self.model.foreach(_update_other_song)
                _update_warnings()
            elif old_path["unique"] and new_path["duplicate"]:
                _make_duplicate(entry, True)
                entry.export_path = entered_path
            elif old_path["unique"] and new_path["unique"]:
                entry.export_path = entered_path
                self.model.foreach(_update_other_song)

            # Update the model and the summary field
            self.model.set_row(self.model.get_iter(path), self._make_model_row(entry))
            self._update_preview_summary()

    def _update_model_value(self, iter_, column, value):
        """
        Set the data in a since cell of the ListStore model.

        :param iter_:  A Gtk.TreeIter for the row being modified.
        :param column: The name of the column to modify.
        :param value:  The new value for the cell.
        """
        self.model.set_value(iter_, self._model_col_id(column), value)

    def _model_col_id(self, name):
        """
        Get the column ID from the given name.

        :param name: The column name to search for.
        :raises: KeyError if a column with the given name does not exist.
        """
        return self.model_cols[name][0]

    @staticmethod
    def _make_model_row(entry):
        """
        Create a new row to insert into the ListStore model.

        :param entry: The Entry to insert.
        """
        return [entry, entry.tag, entry.filename, entry.export_path]

    @staticmethod
    def _run_pending_events():
        """
        Prevent the application from becoming unresponsive.
        """
        while Gtk.events_pending():
            Gtk.main_iteration()

    def _start_preview(self, button):
        """
        Start the generation of export paths for all songs.

        :param button: The start preview button.
        """
        print_d(_("Starting synchronization preview"))
        self.running = True

        # Summary labels
        self.status_operation.set_label(_("Synchronization preview in progress."))
        self.status_operation.set_visible(True)
        self.status_progress.set_visible(False)
        self.status_duplicates.set_visible(False)
        self.status_deletions.set_visible(False)

        # Change button visibility
        self.preview_start_button.set_visible(False)
        self.preview_stop_button.set_visible(True)

        self.c_songs_copy = self.c_song_dupes = self.c_songs_delete = 0
        if self._run_preview() is None:
            return

        self._stop_preview()
        self.sync_start_button.set_sensitive(True)
        print_d(_("Finished synchronization preview"))

    def _stop_preview(self, button=None):
        """
        Stop the generation of export paths for all songs.

        :param button: The stop preview button.
        """
        if button:
            print_d(_("Stopping synchronization preview"))
            self.status_operation.set_label(_("Synchronization preview was stopped."))
        else:
            self.status_operation.set_label(_("Synchronization preview has finished."))
        self.status_operation.set_visible(True)
        self.running = False

        # Change button visibility
        self.preview_start_button.set_visible(True)
        self.preview_stop_button.set_visible(False)

        self._update_preview_summary()

    def _run_preview(self):
        """
        Show the export paths for all songs to be synchronized.

        :return: Whether the generation of preview paths was successful.
        """
        destination_path, pattern = self._get_valid_inputs()
        if None in {destination_path, pattern}:
            return False
        self.expanded_destination = os.path.expanduser(destination_path)

        # Get a list containing all songs to export
        songs = self._get_songs_from_queries()
        if not songs:
            return False
        self.model.clear()
        export_paths = []

        for song in songs:
            if not self.running:
                print_d(_("Stopped synchronization preview"))
                return None
            self._run_pending_events()
            if not self.destination_entry.get_text():
                print_d(_("A different plugin was selected - stop preview"))
                return False

            export_path = self._get_export_path(song, destination_path, pattern)
            if not export_path:
                return False

            entry = Entry(song, export_path)

            expanded_path = os.path.expanduser(export_path)
            if expanded_path in export_paths:
                entry.tag = Entry.Tags.SKIP_DUPLICATE
                self.c_song_dupes += 1
            else:
                entry.tag = Entry.Tags.PENDING_COPY
                self.c_songs_copy += 1
                export_paths.append(expanded_path)

            self.model.append(row=self._make_model_row(entry))

        # List files to delete
        for root, __, files in os.walk(self.expanded_destination):
            for name in files:
                file_path = os.path.join(root, name)
                if file_path not in export_paths and "cover.jpg" not in file_path:
                    entry = Entry(None)
                    entry.filename = file_path
                    entry.tag = Entry.Tags.PENDING_DELETE
                    self.model.append(row=self._make_model_row(entry))
                    self.c_songs_delete += 1

        return True

    def _update_preview_summary(self):
        """
        Update the preview summary text field.
        """
        prefix = _("Synchronization will:") + self.summary_sep
        preview_progress = []

        if self.c_songs_copy > 0:
            counter = self.c_songs_copy
            preview_progress.append(
                ngettext(
                    "attempt to write {count} file",
                    "attempt to write {count} files",
                    counter,
                ).format(count=counter)
            )

        if self.c_song_dupes > 0:
            counter = self.c_song_dupes
            preview_progress.append(
                ngettext(
                    "skip {count} duplicate file",
                    "skip {count} duplicate files",
                    counter,
                ).format(count=counter)
            )
            for child in self.status_duplicates.get_children():
                child.set_visible(True)
            self.status_duplicates.set_visible(True)

        if self.c_songs_delete > 0:
            counter = self.c_songs_delete
            preview_progress.append(
                ngettext("delete {count} file", "delete {count} files", counter).format(
                    count=counter
                )
            )
            for child in self.status_deletions.get_children():
                child.set_visible(True)
            self.status_deletions.set_visible(True)

        preview_progress_text = self.summary_sep_list.join(preview_progress)
        if preview_progress_text:
            preview_progress_text = prefix + preview_progress_text
            self.status_progress.set_label(preview_progress_text)
            self.status_progress.set_visible(True)
            print_d(preview_progress_text)

    def _get_paths(self):
        """
        Build a list of all current export paths for the songs to be
        synchronized.
        """
        paths = {}
        for row in self.model:
            entry = row[self._model_col_id("entry")]
            if entry.tag != Entry.Tags.PENDING_DELETE and entry.export_path:
                if entry.export_path not in paths.keys():
                    paths[entry.export_path] = 1
                else:
                    paths[entry.export_path] += 1
        return paths

    def _show_sync_error(self, title, message):
        """
        Show an error message whenever a synchronization error occurs.

        :param title:   The title of the message popup.
        :param message: The error message.
        """
        qltk.ErrorMessage(self.main_vbox, title, message).run()
        print_e(title)

    def _get_valid_inputs(self):
        """
        Ensure that all user inputs have been given. Shows a popup error message
        if values are not as expected.

        :return: The entered destination path and an fsnative pattern,
                 or None if an error occurred.
        """
        # Get text from the destination path entry
        destination_path = self.destination_entry.get_text()
        if not destination_path:
            self._show_sync_error(
                _("No destination path provided"),
                _("Please specify the directory where songs should be exported."),
            )
            return None, None

        # Get text from the export pattern entry
        export_pattern = self.export_pattern_entry.get_text()
        if not export_pattern:
            self._show_sync_error(
                _("No export pattern provided"),
                _(
                    "Please specify an export pattern for the "
                    "names of the exported songs."
                ),
            )
            return None, None

        # Combine destination path and export pattern to form the full pattern
        full_export_path = os.path.join(destination_path, export_pattern)
        try:
            pattern = FileFromPattern(full_export_path)
        except ValueError:
            self._show_sync_error(
                _("Export path is not absolute"),
                _(
                    'The pattern\n\n{}\n\ncontains "/" but does not start '
                    "from root. Please provide an absolute destination path by "
                    "making sure it starts with / or ~/."
                ).format(util.bold(full_export_path)),
            )
            return None, None

        return destination_path, pattern

    def _get_songs_from_queries(self):
        """
        Build a list of songs to be synchronized, filtered using the
        selected saved searches.

        :return: A list of the selected songs.
        """
        enabled_queries = []
        for query_name, query in self.queries.items():
            query_config = self.CONFIG_QUERY_PREFIX + query_name
            if self.config_get_bool(query_config):
                enabled_queries.append(query)

        if not enabled_queries:
            self._show_sync_error(
                _("No saved searches selected"),
                _("Please select at least one saved search."),
            )
            return []

        selected_songs = []
        for song in app.library.itervalues():
            if any(query.search(song) for query in enabled_queries):
                selected_songs.append(song)

        if not selected_songs:
            self._show_sync_error(
                _("No songs in the selected saved searches"),
                _("All selected saved searches are empty."),
            )
            return []

        print_d(_("Found {} songs to synchronize").format(len(selected_songs)))
        return selected_songs

    def _get_export_path(self, song, destination_path, export_pattern):
        """
        Use the given pattern of song tags to build the destination path
        for a song.

        :param song:             The song for which to build the export path.
        :param destination_path: The user-entered destination path.
        :param export_pattern:   An fsnative file path pattern.
        :return: A safe full destination path for the song.
        """
        new_name = Path(export_pattern.format(song))

        try:
            relative_name = new_name.relative_to(self.expanded_destination)
        except ValueError as ex:
            self._show_sync_error(
                _("Mismatch between destination path and export pattern"),
                _(
                    "The export pattern starts with a path that "
                    "differs from the destination path. Please "
                    "correct the pattern.\n\nError:\n{}"
                ).format(ex),
            )
            return None

        return os.path.join(destination_path, self._make_safe_name(relative_name))

    def _make_safe_name(self, input_path):
        """
        Make a file path safe by replacing unsafe characters.

        :param input_path: A relative Path.
        :return: The given path, with any unsafe characters replaced.
                 Returned as a string.
        """
        # Remove diacritics (accents)
        safe_filename = unicodedata.normalize("NFKD", str(input_path))
        safe_filename = "".join(
            c for c in safe_filename if not unicodedata.combining(c)
        )

        if os.name != "nt":
            # Ensure that Win32-incompatible chars are always removed.
            # On Windows, this is called during `FileFromPattern`.
            safe_filename = strip_win32_incompat_from_path(safe_filename)

        return safe_filename

    def _start_sync(self, button):
        """
        Start the song synchronization.

        :param button: The start sync button.
        """
        # Check sort column
        sort_columns = [
            c.get_title()
            for c in self.details_tree.get_columns()
            if c.get_sort_indicator()
        ]
        if "Status" in sort_columns:
            self._show_sync_error(
                _("Unable to sync"),
                _("Cannot start synchronization while sorting by <b>Status</b>."),
            )
            return

        print_d(_("Starting song synchronization"))
        self.running = True

        # Summary labels
        self.status_operation.set_label(_("Synchronization in progress."))
        self.status_duplicates.set_visible(False)
        self.status_deletions.set_visible(False)

        # Change button visibility
        self.sync_start_button.set_visible(False)
        self.sync_stop_button.set_visible(True)

        if not self._run_sync():
            return

        self._stop_sync()
        print_d(_("Finished song synchronization"))

    def _stop_sync(self, button=None):
        """
        Stop the song synchronization.

        :param button: The stop sync button.
        """
        if button:
            print_d(_("Stopping song synchronization"))
            self.status_operation.set_label(_("Synchronization was stopped."))
        else:
            self.status_operation.set_label(_("Synchronization has finished."))

        self.running = False

        # Change button visibility
        self.sync_start_button.set_visible(True)
        self.sync_stop_button.set_visible(False)

    def _run_sync(self):
        """
        Synchronize the songs from the selected saved searches
        with the specified folder.

        :return: Whether the synchronization was successful.
        """
        self.c_files_copy = self.c_files_skip = self.c_files_skip_previous = (
            self.c_files_dupes
        ) = self.c_files_delete = self.c_files_failed = 0
        self.model.foreach(self._sync_entry)
        if not self.running:
            return False
        self._remove_empty_dirs()
        return True

    def _sync_entry(self, model, path, iter_, *data):
        """
        Synchronize a single song.
        This is a callback function passed to Gtk.TreeModel.foreach() to iterate
        over the rows in a tree model.

        :return: True to stop iterating, False to continue.
        """
        entry = model[path][self._model_col_id("entry")]
        if not self.running:
            print_d(_("Stopped song synchronization"))
            return True
        self._run_pending_events()
        if not self.destination_entry.get_text():
            print_d(_("A different plugin was selected - stop synchronization"))
            return True

        print_d(
            _('{tag} - "{filename}"').format(tag=entry.tag, filename=entry.filename)
        )

        if not entry.export_path and not entry.tag:
            return False

        if entry.tag == Entry.Tags.PENDING_COPY:
            # Export, skipping existing files
            expanded_path = os.path.expanduser(entry.export_path)
            if os.path.exists(expanded_path):
                entry.tag = Entry.Tags.RESULT_SKIP_EXISTING
                self._update_model_value(iter_, "tag", entry.tag)
                self.c_files_skip += 1
            else:
                entry.tag = Entry.Tags.IN_PROGRESS_SYNC
                self._update_model_value(iter_, "tag", entry.tag)

                song_folders = os.path.dirname(expanded_path)
                os.makedirs(song_folders, exist_ok=True)
                try:
                    shutil.copyfile(entry.filename, expanded_path)
                except Exception as ex:
                    entry.tag = Entry.Tags.RESULT_FAILURE + ": " + str(ex)
                    self._update_model_value(iter_, "tag", entry.tag)
                    print_exc()
                    self.c_files_failed += 1
                else:
                    entry.tag = Entry.Tags.RESULT_SUCCESS
                    self._update_model_value(iter_, "tag", entry.tag)
                    self.c_files_copy += 1

        elif entry.tag == Entry.Tags.SKIP_DUPLICATE:
            self.c_files_dupes += 1

        elif entry.tag == Entry.Tags.PENDING_DELETE:
            # Delete file
            try:
                entry.tag = Entry.Tags.IN_PROGRESS_DELETE
                self._update_model_value(iter_, "tag", entry.tag)

                os.remove(entry.filename)
            except Exception as ex:
                entry.tag = Entry.Tags.RESULT_FAILURE + ": " + str(ex)
                self._update_model_value(iter_, "tag", entry.tag)
                print_exc()
                self.c_files_failed += 1
            else:
                entry.tag = Entry.Tags.RESULT_SUCCESS
                self._update_model_value(iter_, "tag", entry.tag)
                self.c_files_delete += 1

        else:
            self.c_files_skip_previous += 1

        self._update_sync_summary()
        return False

    def _remove_empty_dirs(self):
        """
        Delete all empty sub-directories from the given path.
        """
        for root, dirs, files in os.walk(self.expanded_destination, topdown=False):
            for dirname in dirs:
                dir_path = os.path.realpath(os.path.join(root, dirname))
                last_file_is_cover = files and files[0] == "cover.jpg"
                if not files or last_file_is_cover:
                    entry = Entry(None)
                    entry.filename = dir_path
                    entry.tag = Entry.Tags.IN_PROGRESS_DELETE
                    iter_ = self.model.append(row=self._make_model_row(entry))
                    print_d(_('Removing "{}"').format(entry.filename))
                    self.c_songs_delete += 1
                    try:
                        if last_file_is_cover:
                            os.remove(os.path.join(dir_path, files[0]))
                        os.rmdir(dir_path)
                    except Exception as ex:
                        entry.tag = Entry.Tags.RESULT_FAILURE + ": " + str(ex)
                        self._update_model_value(iter_, "tag", entry.tag)
                        print_exc()
                        self.c_files_failed += 1
                    else:
                        entry.tag = Entry.Tags.RESULT_SUCCESS
                        self._update_model_value(iter_, "tag", entry.tag)
                        self.c_files_delete += 1
                    self._update_sync_summary()

    def _update_sync_summary(self):
        """
        Update the synchronization summary text field.
        """
        sync_summary_prefix = _("Synchronization has:") + self.summary_sep
        sync_summary = []

        if self.c_files_copy > 0 or self.c_files_skip > 0:
            text = []

            counter = self.c_files_copy
            text.append(
                ngettext(
                    "written {count}/{total} file",
                    "written {count}/{total} files",
                    counter,
                ).format(count=counter, total=self.c_songs_copy)
            )

            if self.c_files_skip > 0:
                counter = self.c_files_skip
                text.append(
                    ngettext(
                        "(skipped {count} existing file)",
                        "(skipped {count} existing files)",
                        counter,
                    ).format(count=counter)
                )

            sync_summary.append(self.summary_sep.join(text))

        if self.c_files_dupes > 0:
            counter = self.c_files_dupes
            sync_summary.append(
                ngettext(
                    "skipped {count}/{total} duplicate file",
                    "skipped {count}/{total} duplicate files",
                    counter,
                ).format(count=counter, total=self.c_song_dupes)
            )

        if self.c_files_delete > 0:
            counter = self.c_files_delete
            sync_summary.append(
                ngettext(
                    "deleted {count}/{total} file",
                    "deleted {count}/{total} files",
                    counter,
                ).format(count=counter, total=self.c_songs_delete)
            )

        if self.c_files_failed > 0:
            counter = self.c_files_failed
            sync_summary.append(
                ngettext(
                    "failed to sync {count} file",
                    "failed to sync {count} files",
                    counter,
                ).format(count=counter)
            )

        if self.c_files_skip_previous > 0:
            counter = self.c_files_skip_previous
            sync_summary.append(
                ngettext(
                    "skipped {count} file synchronized previously",
                    "skipped {count} files synchronized previously",
                    counter,
                ).format(count=counter)
            )

        sync_summary_text = self.summary_sep_list.join(sync_summary)
        sync_summary_text = sync_summary_prefix + sync_summary_text
        self.status_progress.set_label(sync_summary_text)
        print_d(sync_summary_text)



# Copyright 2018 Peter Strulo
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

from gi.repository import Gtk
from quodlibet import _
from quodlibet.plugins.query import QueryPlugin, markup_for_syntax
from quodlibet.plugins import PluginConfigMixin
from quodlibet.qltk import Frame


class MissingQuery(QueryPlugin, PluginConfigMixin):
    PLUGIN_ID = "missing_query"
    PLUGIN_NAME = _("Missing Query")
    PLUGIN_DESC = _("Matches songs without the given tag.")
    key = "missing"
    query_syntax = "@(missing: artist)"
    usage = markup_for_syntax(query_syntax)

    def search(self, data, body):
        val = data.get(body.strip() if body else "", None)
        return val is None or (
            self.config_get_bool("include_empty", True) and val == ""
        )

    @classmethod
    def PluginPreferences(cls, window):
        example = super().PluginPreferences(window)
        box = Gtk.VBox()
        box.pack_start(example, True, True, 0)

        prefs_box = Gtk.VBox()
        button = cls.ConfigCheckButton(_("Include empty tags"), "include_empty", True)
        prefs_box.pack_start(button, False, False, 6)
        frame = Frame(_("Preferences"), child=prefs_box)
        box.pack_start(frame, True, True, 12)
        return box
